use std::net::IpAddr;
use std::time::Duration;
use crate::auxiliary::response_template::{response_error, response_json, response_msg};
use crate::database_connector::Database;
use crate::model::attraction_type::AttractionType;
use crate::model::device::DeviceState;
use crate::model::device_type::DeviceType;
use crate::model::extension::Extension;
use crate::model::launch_configuration::LaunchConfiguration;
use crate::model::controller_alias::ControllerAlias;
use crate::model::session::{Session, SessionState};
use actix_web::web::{Data, Json};
use actix_web::HttpResponse;
use serde::{Serialize, Deserialize};
use uuid::Uuid;
use crate::auxiliary::config_reader::read_config_field;
use super::runner::AttractionReqData;


#[post("/admin/types/device/new")]
pub async fn add_device_type(data: Data<Database>, device_type: Json<DeviceType>) -> HttpResponse {
    let device_type = device_type.clone();
    response_json(
        data.device_type_repo
            .create(DeviceType::new(
                device_type.name,
                device_type.configurations,
                device_type.is_command_remote,
            ))
            .await
            .unwrap(),
    )
}

#[get("/admin/types/device/get_all")]
pub async fn get_device_types(data: Data<Database>) -> HttpResponse {

    response_json(
        data.device_type_repo.get_all().await.unwrap()
    )
}


#[post("/admin/types/attraction/new")]
pub async fn add_attraction_type(
    data: Data<Database>,
    attraction_type: Json<AttractionType>,
) -> HttpResponse {
    let attraction_type = attraction_type.clone();
    response_json(
        data.attraction_type_repo
            .create(attraction_type)
            .await
            .unwrap(),
    )
}

#[get("/get/attractions")]
pub async fn get_attractions(
    data: Data<Database>,
) -> HttpResponse {
    response_json(
        data.attraction_type_repo.get_all()
            .await
            .unwrap(),
    )
}


#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct IpPortData {
    ip: IpAddr,
    port: u16,
}


#[post("/controllers/vibrate")]
pub async fn vibrate_controllers(
    data: Data<Database>, obj: Json<DeviceIdData>
    ) -> HttpResponse {
    let obj = obj.0;
    let device_data = data.device_repo.get_DeviceData_by_uid(&obj.id).await.unwrap();

    let reqwest_client = reqwest::Client::new();
    let max_response_time = read_config_field("max-response-time-ms")
        .unwrap()
        .as_i64()
        .unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    let resp = reqwest_client
        .post(format!("http://{}:{}/controllers/vibrate", &device_data.ip, device_data.port))
        .timeout(max_response_duration)
        .send().await;

    match resp {
        Ok(_) => {
            response_msg("Ok")
        }
        Err(e) => {
            let msg = format!("{:?}", e);
            response_error(HttpResponse::InternalServerError(), msg.as_str())
        }
    }

}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct DeviceStateData {
    id: Uuid,
    state: DeviceState,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct DeviceIdData {
    id: Uuid,
}


#[post("/device/state")]
pub async fn update_device_state(data: Data<Database>, obj: Json<DeviceStateData>) -> HttpResponse {
    let obj = obj.0;
    data.device_repo.update_DeviceData_state_by_id(&obj.id, obj.state).await;
    response_msg("Ok")
}

#[post("/device/update")]
pub async fn update_device(data: Data<Database>, obj: Json<DeviceIdData>) -> HttpResponse {
    let obj = obj.0;
    let device_data = data.device_repo.get_DeviceData_by_uid(&obj.id).await.unwrap();

    let reqwest_client = reqwest::Client::new();
    let max_response_time = read_config_field("max-response-time-ms")
        .unwrap()
        .as_i64()
        .unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    let resp = reqwest_client
        .post(format!("http://{}:{}/update", &device_data.ip, device_data.port))
        .timeout(max_response_duration)
        .send().await;
    //TODO(Handle errors)
    response_msg("Ok")
}


#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct NewConfigurationData {
    device_type_id: Uuid,
    configuration: LaunchConfiguration
}

#[post("/device-type/configuraitons/add")]
pub async fn add_configuration(data: Data<Database>, obj: Json<NewConfigurationData>) -> HttpResponse {
    let obj = obj.0;
    let device_type = data.device_type_repo.get_DeviceType_by_uid(&obj.device_type_id).await.unwrap();
    let mut configurations = device_type.configurations;
    configurations.push(obj.configuration);
    data.device_type_repo.update_DeviceType_configurations_by_uid(&obj.device_type_id, &configurations).await;

    response_msg("Ok")
}

#[post("/extension/add")]
pub async fn add_extension(data: Data<Database>, obj: Json<Extension>) -> HttpResponse {
    let obj = obj.0;
    data.extensions_repo.add_Extension(obj).await.unwrap();
    response_msg("Ok")
}

#[derive(Debug, Deserialize, Serialize)]
pub struct SessionStateReqData {
    session_id: Uuid,
    session_state: SessionState
}

#[post("/admin/session_state")]
async fn update_session_state(req: Json<SessionStateReqData>, data: Data<Database>) -> HttpResponse {
    let session_id = req.session_id;
    let session_state = req.session_state;
    data.session_repo.update_Session_state_by_id(&session_id, session_state).await;
    response_msg("Ok")
}


#[post("/admin/controller_alias")]
pub async fn update_controller_alias(req: Json<ControllerAlias>, data: Data<Database>) -> HttpResponse {
    let obj = req.0;
    let serial_number = obj.serial_number.as_str();
    let alias = obj.alias.as_str();
    let res = data.controller_alias_repo.update_ControllerAlias_alias_by_serial_number(serial_number, alias).await;
    if res.is_ok() { response_msg("Ok") }
    else { response_error(HttpResponse::InternalServerError(), "Unexistent controller") }
}


#[derive(Debug, Deserialize, Serialize)]
struct SessionExistsReqData {
    session_id: Uuid
}


#[derive(Debug, Deserialize, Serialize)]
struct SessionExistsResData {
    result: bool
}

#[post("/admin/session_exists")]
pub async fn check_session_existence(req: Json<SessionExistsReqData>, data: Data<Database>) -> HttpResponse {
    let session_id = req.session_id;
    match data.session_repo.get_Session_by_id(&session_id).await {
        Ok(_) => response_json(SessionExistsResData{result: true}),
        Err(_) => response_json(SessionExistsResData{result: false})
    }
}
