use std::process::Command;
use std::thread;
use std::time::Duration;
use actix_web::HttpResponse;
use actix_web::web::Json;
use ads::{AmsAddr, Error};
use serde::{Serialize, Deserialize};
use crate::auxiliary::config_reader::get_exe_path;
use crate::auxiliary::response_template::{response_error, response_msg};
use crate::obj_to_json_resp;


const MOVE_ID_PARAMETER: &str = "GVL_Axis.nMoveID";
const READY_PARAMETER : &str = "GVL_Axis.bReady";
const START_PARAMETER: &str = "GVL_Axis.StartFromPC";


#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct RunTwinCATScriptData {
    remote_ads: String,
    remote_ip: String
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct TwinCATData {
    remote_ads: String,
    remote_ip: String,
    move_id: i32
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct StartPLCData {
    ams_net_id: String,
    move_id: i32
}

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct StopPlcData {
    ams_net_id: String,
}


pub fn handle_ads_error(e: Error) -> HttpResponse {
    return match e
    {
        Error::Io(e, _) => {response_error(HttpResponse::InternalServerError(), e)}
        Error::Ads(_, e, _) => {response_error(HttpResponse::InternalServerError(), e)}
        Error::Reply(_, e, _) => {response_error(HttpResponse::BadRequest(), e)}
        Error::Overflow(_) => {response_error(HttpResponse::BadRequest(), "A value exceeds the allowed 32 bits for ADS.")}
    }

}

#[post("/ads/run/twin_cat_script")]
pub async fn run_twin_cat_script(req: Json<RunTwinCATScriptData>) -> HttpResponse {
    let exe_dir = get_exe_path();

    let program_file_path = format!("{exe_dir}\\pyads_connect.py");
    let program_file_path = program_file_path.as_str();

    let res = Command::new("python")
        .args([program_file_path, &req.remote_ads, &req.remote_ip])
        .output();
    if let Err(e) = res {
        return response_error(HttpResponse::InternalServerError(), e.to_string().as_str())
    }
    let output = res.unwrap();
    let stdout = output.stdout;
    let status = output.status.code().unwrap_or_else(|| -> i32{0});
    return obj_to_json_resp!(
                 status: i32 = status
            )
}

#[post("/ads/v1/start")]
pub async fn start_ads_script(req: Json<StartPLCData>) -> HttpResponse {
    let exe_dir = get_exe_path();

    let program_file_path = format!("{exe_dir}\\start_plc.py");
    let program_file_path = program_file_path.as_str();

    let res = Command::new("python")
        .args([program_file_path, &req.ams_net_id, &req.move_id.to_string()])
        .output();
    if let Err(e) = res {
        return response_error(HttpResponse::InternalServerError(), e.to_string().as_str())
    }
    let output = res.unwrap();
    let stdout = output.stdout;
    let status = output.status.code().unwrap_or_else(|| -> i32{0});
    return obj_to_json_resp!(
                 status: i32 = status
            )
}

#[post("/ads/v1/stop")]
pub async fn stop_ads_script(req: Json<StopPlcData>) -> HttpResponse {
    let exe_dir = get_exe_path();

    let program_file_path = format!("{exe_dir}\\start_plc.py");
    let program_file_path = program_file_path.as_str();

    let res = Command::new("python")
        .args([program_file_path, &req.ams_net_id])
        .output();
    if let Err(e) = res {
        return response_error(HttpResponse::InternalServerError(), e.to_string().as_str())
    }
    let output = res.unwrap();
    let stdout = output.stdout;
    let status = output.status.code().unwrap_or_else(|| -> i32{0});
    return obj_to_json_resp!(
                 status: i32 = status
            )
}


#[post("/ads/v2/start")]
pub async fn start_ads(req: Json<TwinCATData>) -> HttpResponse {
    let req = req.0.clone();
    let client = ads::Client::new((req.remote_ip, ads::PORT), ads::Timeouts::none(),
                                  ads::Source::Auto);

    if let Err(e) = client {
        return handle_ads_error(e);
    }


    let client = client.unwrap();
    let device = client.device(AmsAddr::new(req.remote_ads.parse().unwrap(),
                                            ads::ports::TC3_PLC_SYSTEM1));

    let move_id_handle = ads::Handle::new(device, MOVE_ID_PARAMETER).unwrap();
    let start_handle = ads::Handle::new(device, START_PARAMETER).unwrap();
    let ready_handle = ads::Handle::new(device, READY_PARAMETER).unwrap();

    if let Err(e) = move_id_handle.write_value(&req.move_id){
        return handle_ads_error(e);
    }

    thread::sleep(Duration::from_secs(1));

    let is_ready: Result<u8,_> = ready_handle.read_value(); //TODO(Check type)
    if let Err(e) = is_ready {
        return handle_ads_error(e);
    }
    let is_ready = is_ready.unwrap();

    if is_ready == 1 {
        let start_move = start_handle.write_value(&1);
        if let Err(e) = start_move {
            return handle_ads_error(e);
        }

    }
    return response_msg("Ok")
}

#[post("/ads/v2/stop")]
pub async fn stop_ads(req: Json<RunTwinCATScriptData>) -> HttpResponse {
    let req = req.0.clone();
    let client = ads::Client::new((req.remote_ip, ads::PORT), ads::Timeouts::none(),
                                  ads::Source::Auto);

    if let Err(e) = client {
        return handle_ads_error(e);
    }


    let client = client.unwrap();
    let device = client.device(AmsAddr::new(req.remote_ads.parse().unwrap(), ads::ports::TC3_PLC_SYSTEM1));

    let start_handle = ads::Handle::new(device, START_PARAMETER).unwrap();

    let stop_move = start_handle.write_value(&0);
    if let Err(e) = stop_move {
        return handle_ads_error(e);
    }

    return response_msg("Ok")
}
