use crate::api::response::Response;
use crate::auxiliary::config_reader::read_ip_port;
use crate::auxiliary::obj_to_json_resp;
use crate::auxiliary::response_template::{response_error, response_json, response_msg};
use crate::database_connector::Database;
use crate::model::device::{DeviceData, DeviceState};
use crate::model::device::DeviceState::{Inactive, Unavailable};
use crate::model::device_system_info::PivotData;
use crate::model::wol_data::WolData;
use crate::{read_config_field, obj_to_json_resp};
use crate::updates::upload_project;
use actix_web::error::ErrorInternalServerError;
use actix_web::web::{Data, Json};
use actix_web::{HttpResponse, HttpRequest};
use either::{Either, Left};
use macaddr::MacAddr6;
use serde::{Deserialize, Serialize};
use ssh2::Session;
use std::ascii::escape_default;
use std::net::{IpAddr, Ipv4Addr, TcpStream, UdpSocket};
use std::str::FromStr;
use std::time::Duration;
use uuid::Uuid;
use crate::auxiliary::id_name_eq::{IdName, IdNameData, validate_find_by_name_result};
use crate::auxiliary::id_name_eq::IdName::Name;

pub type Devices = Response<DeviceData>;

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct NewDeviceRequest {
    ip: IpAddr,
    device_name: Option<String>,
    port: u16,
    device_type: Uuid,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct AddDeviceToGroupData {
    device_id: String,
    group_id: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct AddMacAddr {
    device_id: Uuid,
    mac_addr: MacAddr6,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct WolDataWithIp {
    ip: IpAddr,
    mac_addr: String,
}

#[derive(Debug,Deserialize,Serialize)]
pub struct DeviceId {
    id: Uuid
}

#[derive(Debug, Deserialize, Serialize)]
pub struct DeviceStateData {
    state: DeviceState
}

fn generate_wol_msg(mac_addr: MacAddr6) -> [u8; 102] {
    let mac_addr_bytes = mac_addr.as_bytes();
    let mut data = [0xFF; 102];
    for iter in 1..=16 {
        for i in 0..6 {
            data[iter * 6 + i] = mac_addr_bytes[i];
        }
    }
    data
}



async fn is_device_exists(data: &Data<Database>, ip: IpAddr) -> Either<HttpResponse, ()> {
    let device_data_future = data.device_repo.get_DeviceData_by_ip(&ip);
    if let Ok(_) = device_data_future.await {
        return Left(response_error(
            HttpResponse::BadRequest(),
            format!("Device with ip({ip}) already exists").as_str(),
        ));
    }
    Either::Right(())
}

async fn check_device_type(
    data: &Data<Database>,
    device_id: Uuid,
    device_type_outer: String,
) -> Option<HttpResponse> {
    let device_data_res = data.device_repo.get_DeviceData_by_uid(&device_id).await;

    if let Ok(device_data) = device_data_res {
        let device_type_str = device_data.device_type.to_string();
        let outer_device_type_str = device_type_outer.to_string();
        match device_type_str {
            _ if outer_device_type_str == device_type_str => {}
            _ => {
                return Some(response_error(
                    HttpResponse::BadRequest(),
                    "The device has the wrong device type",
                ))
            }
        }
    } else {
        return Some(response_error(
            HttpResponse::BadRequest(),
            "Device data doesn't exist",
        ));
    }
    return None;
}

#[post("/connections/delete")]
pub async fn delete_device(data: Data<Database>, device: Json<DeviceId>) -> HttpResponse {
    let res: Result<_,_> = data.device_repo.delete_DeviceData_by_id(&device.id).await;
    if let Err(_) = res {
       return response_error(HttpResponse::InternalServerError(), "Can't remove");
    }
    response_msg("Success")
}

#[get("/connections/devices")]
pub async fn list_devices(data: Data<Database>) -> HttpResponse {
    let devices_data = data.device_repo.get_all_DeviceData().await.unwrap();
    let devices = Devices {
        results: devices_data,
    };

    response_json(devices)
}

#[get("/connections/device")]
pub async fn get_device_data(data: Data<Database>, req: Json<IdNameData>) -> HttpResponse {
    match req.get_type(){
        IdName::Id(id) => {response_json(data.device_repo.get_DeviceData_by_uid(&id).await.unwrap())}
        Name(name) => {
            let res = data.device_repo.get_all_DeviceData_by_name(&name).await.unwrap();
            match validate_find_by_name_result(res) {
                Ok(d) => {response_json(d)},
                Err(r) => {response_error(HttpResponse::BadRequest(), r.as_str())},
            }
        }
        IdName::Error => {response_error(HttpResponse::BadRequest(), "Invalid body")}
    }
}

#[post("/connections/state")]
pub async fn update_state(data: Data<Database>, req: HttpRequest, state: Json<DeviceStateData>) -> HttpResponse {
    let state = state.0.state;
    let peer = req.peer_addr();
    if let None = peer {
        log::error!("Can't retreive peer address");
        return response_error(HttpResponse::InternalServerError(), "Can't retreive peer address");
    }
    let peer = peer.unwrap();
    let ip = peer.ip();
    let device_data = data.device_repo.get_DeviceData_by_ip(&ip).await;
    if let Err(_) = device_data {
        log::error!("Can't get device from ip");
        return response_error(HttpResponse::InternalServerError(), "Can't get device from ip");
    }
    let device_data = device_data.unwrap();

    let res = data.device_repo.update_DeviceData_state_by_id(&device_data.id, state).await;
    if let Err(_) = res {
        log::error!("Can't update state by id");
        return response_error(HttpResponse::InternalServerError(), "Can't update state");
    }

    return response_msg("Ok")
}

#[post("/connections/new/no-check")]
pub async fn connect_without_check(
    data: Data<Database>,
    req: Json<NewDeviceRequest>,
) -> HttpResponse {
    let ip = req.ip;
    let port = req.port;
    let device_type = req.clone().device_type;
    let device_name = req.clone().device_name;

    let device = DeviceData::new(device_type, device_name, ip, port, None);

    let new_device_result = data.device_repo.add_DeviceData(device).await;

    match new_device_result {
        Err(err) => response_error(
            HttpResponse::InternalServerError(),
            err.to_string().as_str(),
        ),
        Ok(res) => response_json(res),
    }
}

#[post("/connections/new")]
pub async fn connect(data: Data<Database>, req: Json<NewDeviceRequest>) -> HttpResponse {
    let ip = req.ip;
    let port = req.port;
    let device_type = req.clone().device_type;
    let device_name = req.clone().device_name;

    let is_device_exists_fut = is_device_exists(&data, ip);
    let is_device_type_exists_fut = data.device_type_repo.exists(device_type);

    let device = DeviceData::new(device_type, device_name, ip, port, None);

    let reqwest_client = reqwest::blocking::Client::new();
    let max_response_time = read_config_field("max-response-time-ms")
        .unwrap()
        .as_i64()
        .unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    let resp = reqwest_client
        .get(format!("http://{ip}:{port}/health_check"))
        .timeout(max_response_duration)
        .send();

    return if let Err(err) = resp {
        response_error(
            HttpResponse::InternalServerError(),
            err.to_string().as_str(),
        )
    } else {
        if let Left(it) = is_device_exists_fut.await {
            return it;
        }
        if let Ok(it) = is_device_type_exists_fut.await {
           if it == false {
               let types = data.device_type_repo.get_all().await.unwrap().iter()
                   .map(|it| (&it.uid, &it.name))
                   .map(|(uid,name)| format!("{uid}: {name},"))
                   .reduce(|p, c| p+c.as_str()).unwrap_or_default();

               return response_error(HttpResponse::BadRequest(),
                                     format!("Invalid device type. Only ({types}) available").as_str())
           }
        }
        else{
            return response_error(HttpResponse::InternalServerError(), "")
        }

        let new_device_result = data.device_repo.add_DeviceData(device).await;

        match new_device_result {
            Err(err) => response_error(
                HttpResponse::InternalServerError(),
                err.to_string().as_str(),
            ),
            Ok(res) => response_json(res),
        }
    };
}

#[post("/connections/new/ssh")]
pub async fn connect_via_ssh(data: Data<Database>, req: Json<NewDeviceRequest>) -> HttpResponse {
    let ssh_port: i64 = dotenv!("SSH_PORT").parse().unwrap();

    //TODO(Check id exist)
    let ip = req.ip;
    let tcp = TcpStream::connect(format!("{ip}:{ssh_port}")).unwrap();
    let mut sess = Session::new().unwrap();
    sess.set_tcp_stream(tcp);
    sess.handshake().unwrap();

    let ssh_user = dotenv!("SSH_USER");
    let ssh_pass = dotenv!("SSH_PASS");

    let auth_result = sess.userauth_password(ssh_user, ssh_pass);
    if auth_result.is_err() {
        let err = auth_result.unwrap_err().to_string();
        return response_error(
            HttpResponse::InternalServerError(),
            format!("Ssh authorization error: {err}").as_str(),
        );
    }

    if let Err(res) = upload_project(&sess) {
        return response_error(HttpResponse::InternalServerError(), res.as_str());
    }
    let _ = sess.disconnect(None, "", None);
    //TODO(Add to db)
    response_msg("Success")
}

#[post("/connections/check")]
pub async fn refresh_device_connection_state(
    data: Data<Database>,
    req: Json<IdNameData>,
) -> HttpResponse {
    let device_id;
    match req.get_type() {
        IdName::Id(id) => {device_id = id}
        Name(_) => {return response_error(HttpResponse::InternalServerError(), "Feature in todo list")}
        IdName::Error => {return response_error(HttpResponse::BadRequest(), "Invalid body")}
    }

    let device_data = data.device_repo.get_DeviceData_by_uid(&device_id).await.unwrap();

    let ip = device_data.ip;
    let port = device_data.port;

    let max_response_time = read_config_field("max-response-time-ms")
        .unwrap()
        .as_i64()
        .unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    let ip_c = ip.clone();
    let port_c = port.clone();

    let reqwest_client = reqwest::Client::new();

    #[derive(Deserialize)]
    struct ResObj {
        state: DeviceState
    }

    let res = reqwest_client
        .get(format!("http://{ip_c}:{port_c}/state"))
        .timeout(max_response_duration)
        .send().await;
    if let Ok(x) = res {
        let response = x.json::<ResObj>().await;
        if let Ok(rem_obj_state) = response {
            let rem_state = rem_obj_state.state;
            let prev_data = data.device_repo.get_DeviceData_by_uid(&device_id).await.unwrap();
            let prev_state= prev_data.device_state;
            let new_state: DeviceState = match rem_state {
                Inactive => Inactive,
                DeviceState::Working  => if prev_state == DeviceState::Prepared {DeviceState::Prepared} else {DeviceState::Working},
                x => x
            };

            data.device_repo
                .update_DeviceData_state_by_id(&device_id, new_state)
                .await;

           if let None = prev_data.mac {
               if let Ok(all_data)  = PivotData::retreive_all_data(format!("{ip_c}:{port_c}"), data.clone()).await {
                   data.device_repo.update_DeviceData_mac_by_id(&device_id, all_data.system_data.mac_addr).await;
               }
           } 
        }
    }
    else {
        data.device_repo
            .update_DeviceData_state_by_id(&device_id, Unavailable)
            .await;
    }

    let device_data = data.device_repo.get_DeviceData_by_uid(&device_id).await.unwrap();
    response_json(device_data)
}

#[post("/connections/check/ip")]
pub async fn refresh_device_connection_state_by_ip(
    data: Data<Database>,
    req: HttpRequest,
    obj: Json<DeviceStateData>
) -> HttpResponse {
    let ip = req.peer_addr().unwrap().ip();

    let device_data = data.device_repo.get_DeviceData_by_ip(&ip).await;
    if let Err(e) = device_data {
        log::warn!("Unknown device({}) trying to change connection: {}", ip, e);
        return response_error(HttpResponse::NotAcceptable(), "Unauthorized device");
    }
    let device_id = device_data.unwrap().id;

    data.device_repo
        .update_DeviceData_state_by_id(&device_id, obj.state)
        .await;

    response_msg("Ok")
}

#[get("/wol/add")]
pub async fn add_wol_device(data: Data<Database>, req: Json<AddMacAddr>) -> HttpResponse {
    let device_id = &req.device_id;
    let mac_addr = &req.mac_addr;
    let wol_data = WolData {
        device_id: device_id.clone(),
        mac_addr: *mac_addr,
    };
    let device_data = data.device_repo.get_DeviceData_by_uid(device_id);

    let wol_device_data = data.wol_repo.get_device_mac_addr(device_id).await;
    if wol_device_data.is_ok() {
        return response_error(HttpResponse::BadRequest(), "Device already exists");
    }
    if let Err(it) = device_data.await {
        return response_error(HttpResponse::BadRequest(), it);
    }

    let device_data = data.wol_repo.add_new_device(wol_data).await.unwrap();
    response_json(device_data)
}

#[post("/wol/wake")]
pub async fn wake_device_up(data: Data<Database>, req: Json<IdNameData>) -> HttpResponse {
    let device_id;
    match req.get_type() {
        IdName::Id(id) => {device_id = id}
        Name(_) => {return response_error(HttpResponse::InternalServerError(), "Feature in todo list")}
        IdName::Error => {return response_error(HttpResponse::BadRequest(), "Invalid body")}
    }
    let device_data = data.device_repo.get_DeviceData_by_uid(&device_id).await;
    if let Err(_) = device_data {
        return response_error(HttpResponse::InternalServerError() ,"Can't retreive device data");
    }
    let device_data = device_data.unwrap();
    if let None = device_data.mac {
        return response_error(HttpResponse::InternalServerError(), "Can't find mac addr for device");
    }

    let mac = MacAddr6::from_str(device_data.mac.unwrap().as_str()).unwrap(); 
    let msg = generate_wol_msg(mac);

    let from = (Ipv4Addr::new(0, 0, 0, 0), 0);
    let to = (Ipv4Addr::new(255, 255, 255, 255), 9);

    let socket = UdpSocket::bind(from).expect("couldn't bind to address");

    let _ = socket.set_broadcast(true);
    let res = socket.send_to(&msg, to).expect("couldn't send data");
    println!("{}", res);

    response_msg("Ok")
}

pub fn show_buf<B: AsRef<[u8]>>(buf: B) -> String {
    String::from_utf8(
        buf.as_ref()
            .iter()
            .map(|b| escape_default(*b))
            .flatten()
            .collect(),
    )
    .unwrap()
}

#[post("/wol/wake/ip")]
pub async fn wake_device_up_with_ip(req: Json<WolDataWithIp>) -> HttpResponse {
    let ip_addr = &req.ip;
    let mac_addr = MacAddr6::from_str(&req.mac_addr).unwrap();
    let msg = generate_wol_msg(mac_addr);
    for e in msg {
        let l = e / 16;
        let r = e % 16;
    }
    let res = show_buf(msg);
    println!("{}", res);

    let from = (Ipv4Addr::new(0, 0, 0, 0), 0);
    let to = (Ipv4Addr::new(255, 255, 255, 255), 9);

    let socket = UdpSocket::bind(from).expect("couldn't bind to address");

    let _ = socket.set_broadcast(true);
    let res = socket.send_to(&msg, to).expect("couldn't send data");
    println!("{}", res);

    response_msg("Ok")
}


