use actix_web::{HttpResponse, web};
use actix_web::web::Data;

use uuid::Uuid;
use std::collections::HashMap;
use crate::database_connector::Database;
use crate::model::device_system_info::PivotData;
use askama::Template;
use crate::auxiliary::config_reader::read_ip_port;
use crate::model::attraction_type::AttractionType;
use crate::model::device::{DeviceData, DeviceState};
use crate::model::device_type::DeviceType;
use crate::model::launch_configuration::LaunchConfiguration;
use crate::model::session::Session;


#[derive(Debug, Clone)]
struct TypedDeviceData {
    device_type: DeviceType,
    devices: Vec<DeviceData>
}

#[derive(Template)]
#[template(path="index.html")]
struct DevicesDataTemplate<'a> {
    ip_port: &'a String,
    devices_data: &'a Vec<DeviceData>,
    attraction_types: &'a Vec<AttractionType>,
    sessions_data: &'a Vec<Session>,
    device_type_names: &'a HashMap<Uuid, String>,
    attraction_type_names: &'a HashMap<Uuid, String>
}

#[derive(Template)]
#[template(path="run_configuration.html")]
struct ConfigurationsDataTemplate<'a> {
    device_id: &'a Uuid,
    configurations_data: &'a Vec<LaunchConfiguration>,
}


#[derive(Template)]
#[template(path="new_device.html")]
struct NewDeviceTemplate<'a> {
    ip_port: &'a String,
    device_types: &'a Vec<DeviceType>
}

#[derive(Template)]
#[template(path="new_attraction.html")]
struct NewAttractionTemplate<'a> {
    ip_port: &'a String,
    device_types: &'a Vec<DeviceType>
}



#[derive(Template)]
#[template(path="run_device_with_parameters.html")]
struct RunDeviceWithParameters<'a> {
    device_id: &'a Uuid,
    configuration_id: &'a Uuid
}

#[derive(Template)]
#[template(path="run_attraction.html")]
struct RunAttractionDataTemplate<'a> {
    attraction_id: &'a Uuid,
    typed_devices_data: &'a Vec<TypedDeviceData>
}

#[derive(Template)]
#[template(path="device_info.html")]
struct ShowPivotDataTemplate<'a> {
    data: &'a Vec<PivotData>
}

#[derive(Clone, Debug)]
struct DeviceDataTemplate<'a> {
    id: &'a str,
    device_type: &'a str,
    attraction_type: &'a str,
    device_state: &'a str,
    ip: &'a str,
    port: &'a str
}

#[derive(Template)]
#[template(path="device.html")]
struct DeviceAdmin<'a> {
    device_id: &'a Uuid,
    configurations: &'a Vec<LaunchConfiguration>
}

#[derive(Template)]
#[template(path="update.html")]
struct UpdateTemplate {
}


#[get("/admin")]
pub async fn admin(data: Data<Database>) -> HttpResponse {
    let devices_data = data.device_repo.get_all_DeviceData().await.unwrap();
    let attraction_types = data.attraction_type_repo.get_all().await.unwrap();
    let mut sessions = data.session_repo.get_all_Session().await.unwrap();
    let ip_port = read_ip_port();

    let device_types = data.device_type_repo.get_all().await.unwrap();
    let mut device_type_names = std::collections::HashMap::new();
    for device_type in &device_types {
        device_type_names.insert(device_type.uid, device_type.name.clone());
    }

    let mut attraction_type_names = std::collections::HashMap::new();
    for attraction_type in &attraction_types {
        attraction_type_names.insert(attraction_type.uid, attraction_type.name.clone());
    }

    sessions.reverse();
    let r = DevicesDataTemplate {
        ip_port: &ip_port,
        devices_data: &devices_data,
        attraction_types: &attraction_types,
        sessions_data: &sessions,
        device_type_names: &device_type_names,
        attraction_type_names: &attraction_type_names,
    };
    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )
}

#[get("/devices")]
pub async fn get_all_devices_data(data: Data<Database>) -> HttpResponse {
    let devices = data.device_repo.get_all_DeviceData().await.unwrap();
    let hosts: Vec<String> = devices.iter().map(|it| format!("{}:{}", it.ip, it.port)).collect();
    
    let data: Vec<_> = hosts.iter().map(|it| PivotData::retreive_all_data(it.clone(), data.clone())).collect();
    let mut data = futures::future::join_all(data).await;
    data.retain(|it| it.is_ok());
    log::info!("Length arr: {}", data.len());
    let data: Vec<_> = data.into_iter().map(|it| it.unwrap()).collect();

    let template = ShowPivotDataTemplate {
        data: &data
    };
    let x = template.render().unwrap();

    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )
}

#[get("/run/{id}")]
pub async fn run_app(path: web::Path<(Uuid,)>, data: Data<Database>) -> HttpResponse {
    let device_types = data.device_type_repo.get_all().await.unwrap();
    let id = path.into_inner().0;

    let device = data.device_repo.get_DeviceData_by_uid(&id).await.unwrap();

    let device_type = device_types.iter().find(|it| {
        println!("{}, {}:", it.uid, id);
        it.uid == device.device_type
    }).unwrap();
    let configurations_data = &device_type.configurations;

    let r = ConfigurationsDataTemplate {
        device_id: &id,
        configurations_data
    };
    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )

}
#[get("/run/attraction/{id}")]
pub async fn run_attraction(path: web::Path<(Uuid,)>, data: Data<Database>) -> HttpResponse {
    let id = path.into_inner().0;
    let attraction_type = data.attraction_type_repo.get(&id).await.unwrap();
    let possible_devices = attraction_type.minimal_necessary_devices;

    let mut typed_devices_data = vec![];
    let device_types_id = possible_devices.keys();
    for i in device_types_id {
        let device_type = data.device_type_repo.get(i).await.unwrap();
        let devices = data.device_repo.get_all_DeviceData_by_device_type_and_state(i, DeviceState::Inactive).await.unwrap();

        let typed_device_data = TypedDeviceData {
            device_type,
            devices
        };
        typed_devices_data.push(typed_device_data);
    }

    let template = RunAttractionDataTemplate {
        attraction_id: &id,
        typed_devices_data: &typed_devices_data
    };
    let template = template.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        template
    )
}



#[get("/new_device")]
pub async fn new_device(data: Data<Database>) -> HttpResponse {
    let device_types = data.device_type_repo.get_all().await.unwrap();
    let ip_port = read_ip_port();
    let r = NewDeviceTemplate {
        ip_port: &ip_port,
        device_types: &device_types
    };

    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )
}

#[get("/new_attraction")]
pub async fn new_attraction(data: Data<Database>) -> HttpResponse {
    let device_types = data.device_type_repo.get_all().await.unwrap();
    let ip_port = read_ip_port();
    let r = NewAttractionTemplate {
        ip_port: &ip_port,
        device_types: &device_types
    };

    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )
}

#[get("/run/device/with_parameters/{id}/{configuration_id}")]
pub async fn run_device_with_parameters(path: web::Path<(Uuid,Uuid)>, data: Data<Database>) -> HttpResponse {
    let (id, configuration) = path.into_inner();

    let r = RunDeviceWithParameters  {
        device_id: &id,
        configuration_id: &configuration
    };

    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )
}

#[get("/device/{id}")]
pub async fn get_device_admin_page(path: web::Path<Uuid>, data: Data<Database>) -> HttpResponse {
    let id = path.into_inner();
    let device_data = data.device_repo.get_DeviceData_by_uid(&id).await.unwrap();
    let device_type = data.device_type_repo.get_DeviceType_by_uid(&device_data.device_type).await.unwrap();

    let r = DeviceAdmin  {
        device_id: &id,
        configurations: &device_type.configurations
    };

    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(
        x
    )
}


#[get("/update")]
pub async fn update_admin_page(data: Data<Database>) -> HttpResponse {
    let r = UpdateTemplate {};
    let x = r.render().unwrap();
    HttpResponse::Ok().content_type("text/html; charset=utf-8").body(x)
}