use std::collections::{HashMap, HashSet};
use crate::auxiliary::response_template::{response_error, response_json, response_msg};
use crate::database_connector::Database;
use crate::model::extension::ExtensionPoint;
use crate::model::sync_data::SyncData;
use actix_web::web::{Data, Json};
use actix_web::HttpResponse;
use futures::future::join_all;
use js_sandbox::Script;
use reqwest::StatusCode;
use serde::{Deserialize, Serialize};
use crate::{read_config_field, obj_to_json_resp};
use std::time::Duration;
use chrono::Utc;
use uuid::Uuid;
use reqwest::header::CONTENT_TYPE;
use crate::auxiliary::id_name_eq::{IdName, IdNameData, validate_find_by_name_result};
use crate::model::attraction_type::AttractionType;
use crate::model::device::{DeviceData, DeviceState};
use crate::model::device_type::DeviceType;
use crate::model::launch_configuration::{LaunchConfiguration, LaunchConfigurationClient};
use crate::model::session::{Session, SessionState};

type DeviceArguments =(IdNameData, Vec<String>);
type DeviceEnvironment = (IdNameData, HashMap<String,String>);


#[derive(Debug, Deserialize, Serialize)]
pub struct DeviceReqData {
    identifier: IdNameData,
    configuration: IdNameData,
    args: Vec<String>,
    env: HashMap<String,String>
}


#[derive(Debug, Deserialize, Serialize)]
pub struct AttractionReqData {
    attraction: IdNameData,
    device_ids: Vec<IdNameData>,
    args: Vec<DeviceArguments>,
    environment: Vec<DeviceEnvironment>
}

#[derive(Debug, Deserialize, Serialize)]
pub struct SessionReqData {
    session_id: Uuid
}
#[derive(Debug, Deserialize, Serialize)]
pub struct SessionResData {
    uid: Uuid
}

fn post_request<T: Serialize>(req: &str, obj: T) -> Result<(StatusCode, String), String>{
    let configuration_serialized = serde_json::to_string(&obj).unwrap();

    let reqwest_client = reqwest::blocking::Client::new();
    let max_response_time = read_config_field("max-response-time-ms")
        .unwrap()
        .as_i64()
        .unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    let resp = reqwest_client
        .post(req)
        .header(CONTENT_TYPE ,"application/json")
        .body(configuration_serialized)
        .timeout(max_response_duration)
        .send();
    if let Err(resp) = resp {
        return  Err(resp.to_string());
    }
    let resp = resp.unwrap();
    let status = resp.status();
    let body = resp.text().unwrap();
    
    Ok((status, body))
}

async fn find_devices_with_type(data: &Data<Database>, device_type: &DeviceType, cnt: i32) -> Result<Vec<DeviceData>, String> {
    let res = data.device_repo.get_n_DeviceData_by_device_type_and_state(cnt, &device_type.uid, DeviceState::Inactive).await;
    match res {
        Ok(d) => Ok(d),
        Err(e) => Err(e.to_string())
    }
}

async fn get_device_data_by_identifier(data: &Data<Database>, identifier: &IdName) -> Result<DeviceData, String> {
    match identifier {
        IdName::Id(id) => {
            Ok(data.device_repo.get_DeviceData_by_uid(&id).await.unwrap())
        }
        IdName::Name(name) => {
            match validate_find_by_name_result(data.device_repo.get_all_DeviceData_by_name(name.as_str()).await.unwrap()) {
                Ok(d) => {
                    Ok(d)
                },
                Err(r) => {
                    Err(r)
                }
            }
        }
        IdName::Error => {Err("Invalid type".to_string())}
    }
}

async fn get_attraction_type_by_identifier(data: &Data<Database>, identifier: &IdName) -> Result<AttractionType, String> {
    let attraction_type;
    match identifier {
        IdName::Id(id) => {
            attraction_type = data.attraction_type_repo.get(&id).await.unwrap();
        }
        IdName::Name(name) => {
            match validate_find_by_name_result(data.attraction_type_repo.get_all_AttractionType_by_name(name.as_str()).await.unwrap()) {
                Ok(d) => {
                    attraction_type = d;
                },
                Err(e) => {return Err(e)}
            }
        }
        IdName::Error => {return Err("Invalid body".to_string())}
    }
    Ok(attraction_type)
}

async fn get_device_type_by_identifier(data: &Data<Database>, identifier: &IdName) -> Result<DeviceType, String> {
    let device_type: DeviceType;
    match identifier {
        IdName::Id(id) => {
            device_type = data.device_type_repo.get(&id).await.unwrap();
        }
        IdName::Name(name) => {
            match validate_find_by_name_result(data.device_type_repo.get_all_DeviceType_by_name(name.as_str()).await.unwrap()) {
                Ok(d) => { device_type = d; },
                Err(e) => {return Err(e)}
            }

        }
        IdName::Error => {return Err("Can't identify additional arguments for device type. Might be missing id or name.".to_string())}
    }
    Ok(device_type)
}

fn get_configuration_by_identifier(device_type: &DeviceType, identifier: &IdName) -> Result<LaunchConfiguration, String> {
    let configuration;
    match identifier {
        IdName::Id(id) => {
            configuration = device_type.configurations.iter()
                .find(|it| it.uid == *id)
                .unwrap();

        }
        IdName::Name(name) => {
            configuration = device_type.configurations.iter()
                .find(|it| it.name == *name)
                .unwrap();
        }
        IdName::Error => {return Err("Invalid body".to_string())}
    }
    Ok(configuration.clone())
}

async fn device_req_data_to_configuration_and_data(data: &Data<Database>, device_req: &DeviceReqData) -> Result<(DeviceData, LaunchConfiguration), String> {
    let device_data = get_device_data_by_identifier(&data, &device_req.identifier.get_type()).await;
    if let Err(e) = device_data {
        return Err(e)
    }
    let device_data = device_data.unwrap();

    let device_type = data.device_type_repo.get(&device_data.device_type).await.unwrap();

    let configuration =
        get_configuration_by_identifier(&device_type, &device_req.configuration.get_type());

    if let Err(e) = configuration {
        return Err(e);
    }

    let configuration = configuration.unwrap().add_arguments(&device_req.args);
    let configuration = configuration.add_env(&device_req.env);
    Ok((device_data, configuration))
}

async fn device_data_and_uid_to_configuration_and_data(data: &Data<Database>, device_data: DeviceData, configuration_uid: &Uuid, args: &Vec<String>, env: &HashMap<String,String>) -> Result<(DeviceData, LaunchConfiguration), String> {

    let device_type = data.device_type_repo.get(&device_data.device_type).await.unwrap();

    let configuration = device_type.configurations.iter().find(|it| it.uid == *configuration_uid);
    if let None = configuration {
        return Err("Can't found configuration".to_string());
    }

    let configuration = configuration.unwrap().add_arguments(&args);
    let configuration = configuration.add_env(&env);
    Ok((device_data, configuration))
}

async fn run_device_configuration_inner(data: &Data<Database>, device_data: &DeviceData, configuration: &LaunchConfiguration)
    -> Result<(StatusCode, String), String> {
    let sync_necessary_data = if configuration.need_sync {Some((data, device_data.id))} else {None};

    let launch_configuration_req = LaunchConfigurationClient::from_launch_configuration
        (configuration,sync_necessary_data).await.unwrap();

    let ip = device_data.ip;
    let port = device_data.port;

    let req = format!("http://{ip}:{port}/launch_app");
    log::info!("Sending configuration to {}", req);
    let blocking_operation = actix_web::web::block(move|| post_request(req.as_str(), launch_configuration_req)).await;

    if let Ok(op_res) = blocking_operation {
        return op_res;
    }
    Err("Can't acquire mutex for post request".to_string())
}

async fn run_device_configuration_req(data: Data<Database>, device_req: &DeviceReqData) -> Result<(StatusCode, String), String> {
    let res = device_req_data_to_configuration_and_data(&data, &device_req).await;
    match res {
        Ok((d,c)) => {run_device_configuration_inner(&data, &d, &c).await}
        Err(e) => {Err(e)}
    }
}

async fn stop_device(
    data: Data<Database>,
    session_id: Uuid,
    device_id: Uuid,
) -> Result<(StatusCode, String), String> {
    let device: DeviceData = data
        .device_repo
        .get_DeviceData_by_uid(&device_id)
        .await
        .unwrap();
    let url = format!("http://{}:{}/interrupt_app/session", device.ip, device.port);
    let obj = SessionResData { uid: session_id };
    let task = move || post_request(url.as_str(), obj);
    actix_web::web::block(task).await.unwrap()
}

#[post("/runner/stop/attraction")]
pub async fn stop_session(
    data: Data<Database>,
    session: Json<SessionReqData>
) -> HttpResponse {
    let session_id: Uuid = session.0.session_id;
    let session: Session = data.session_repo.get_Session_by_id(&session_id).await.unwrap();
    for device_id in session.devices {
        if let Ok((status_code, _)) = stop_device(data.clone(), session_id, device_id).await {
            if status_code == StatusCode::OK {
                data.device_repo.update_DeviceData_state_by_id(&device_id, DeviceState::Inactive).await;
            }
        }
    }
    data.session_repo.update_Session_state_by_id(&session_id, SessionState::Aborted).await;

    response_msg("Ok")
}

#[post("/runner/run/device")]
pub async fn run_device_configuration(
    data: Data<Database>,
    device_data: Json<DeviceReqData>,
) -> HttpResponse {
    log::info!("Running device congifuration: {:?}", device_data);
    let device_data = device_data.0;
    let res =  run_device_configuration_req(data, &device_data).await;
    match res {
        Ok((s,b)) => {obj_to_json_resp!(
               status: u16 = s.as_u16(),
               body: String = b
                )},
        Err(e) => {
            response_error(HttpResponse::InternalServerError(), e.as_str())
        }
    }
}

#[post("/runner/run/attraction")]
pub async fn run_attraction_configuration(
    data: Data<Database>,
    attraction_data: Json<AttractionReqData>,
) -> HttpResponse {
    log::info!("Running attraction: {:?}", attraction_data);

    let extensions = data.extensions_repo.get_all_Extension_by_extension_point(ExtensionPoint::TransformRequestBeforeAtraction).await;
    if let Ok(exts) = extensions {
        //todo
    }

    //Get attraction type
    let attraction_type = match get_attraction_type_by_identifier(&data, &attraction_data.attraction.get_type()).await{
        Ok(attraction_type) => attraction_type,

        Err(e) => return response_error(HttpResponse::BadRequest(), e.as_str())
    };


    //Verify arguments
    let mut devices_id_set = HashSet::new();
    let mut args = vec![];
    let mut env = vec![];

    for (l,r) in &attraction_data.args {
        let device_type = get_device_type_by_identifier(&data, &l.get_type()).await;
        if let Err(e) = device_type {
            return response_error(HttpResponse::BadRequest(), e.as_str());
        }
        let device_type = device_type.unwrap();
        if !devices_id_set.insert(device_type.uid) {
            return response_error(HttpResponse::BadRequest(), "Multiple additional arguments configurations")
        }

        args.push((device_type, r))
    }

    for (l,r) in &attraction_data.environment{
        let device_type = get_device_type_by_identifier(&data, &l.get_type()).await;
        if let Err(e) = device_type {
            return response_error(HttpResponse::BadRequest(), e.as_str());
        }
        let device_type = device_type.unwrap();
        if !devices_id_set.insert(device_type.uid) {
            return response_error(HttpResponse::BadRequest(), "Multiple additional arguments configurations")
        }

        env.push((device_type,r))
    }


    //Finding minimal necessary devices set
    let mut configurations: Vec<(DeviceData, LaunchConfiguration)> = Vec::new();

    let mut minimal_necessary_devices = attraction_type.minimal_necessary_devices.clone();

    for identifier in &attraction_data.device_ids {
        let device = match get_device_data_by_identifier(&data, &identifier.get_type()).await {
            Ok(device) => {
                if device.device_state == DeviceState::Inactive { device }
                else { return response_error(HttpResponse::BadRequest(), "Some device is unavailable or working"); }
            }
            Err(e) => {
                return response_error(HttpResponse::BadRequest(), e.as_str());
            }
        };

        let device_type = data.device_type_repo.get(&device.device_type).await.unwrap();

        if minimal_necessary_devices.contains_key(&device_type.uid){
             if minimal_necessary_devices[&device_type.uid] <= 1 {
                 minimal_necessary_devices.remove(&device_type.uid);
             }
            else {
                *minimal_necessary_devices.get_mut(&device_type.uid).unwrap() -= 1;
            }
        }


        let configuration = attraction_type.configurations.get(&device_type.uid);

        let additional_args = args.iter().find(|(l,_)| l.uid == device_type.uid);
        let additional_env = env.iter().find(|(l,_)| l.uid == device_type.uid);


        let run_configuration = DeviceReqData {
            identifier: IdNameData::new_id(device.id),
            configuration: IdNameData::new_id(*configuration.unwrap()),
            args: if additional_args.is_some() { additional_args.unwrap().1.clone() } else { vec![] },
            env:  if additional_env.is_some() { additional_env.unwrap().1.clone() } else { HashMap::new() },
        };
        let res = device_req_data_to_configuration_and_data(&data, &run_configuration).await;
        if let Err(e) = res {
            return response_error(HttpResponse::BadRequest(), e.as_str());
        }
        configurations.push(res.unwrap());
    }

    let minimal_necessary_devices: Vec<(Uuid, i32)>  = minimal_necessary_devices.into_iter()
        .filter(|(_,cnt)| *cnt > 0).collect();


    //Finding instances for necessary devices
    for (uid, count) in minimal_necessary_devices {
        log::info!("Finding {} device type with uid {}", count, uid);
        let device_type = data.device_type_repo.get(&uid).await.unwrap();

        let configuration = attraction_type.configurations.get(&device_type.uid);
        if let None = configuration {
           return response_error(HttpResponse::BadRequest(), "Can't find configuration");
        }
        let configuration = configuration.unwrap();
        let additional_args = args.iter().find(|(l,_)| l.uid == device_type.uid);
        let additional_env = env.iter().find(|(l,_)| l.uid == device_type.uid);

        let devices = find_devices_with_type(&data, &device_type, count).await;
        if let Err(e) = devices {
            return response_error(HttpResponse::BadRequest(), e.as_str());
        }
        let devices = devices.unwrap();

        let temp_val_args = (device_type.clone(), &vec![]);
        let temp_val_env = (device_type.clone(), &HashMap::new());

        let results = futures::future::join_all(
        devices.into_iter().map(|r|
            device_data_and_uid_to_configuration_and_data(&data, r, configuration,
                                                          additional_args.unwrap_or_else(|| &temp_val_args).1,
                                                          additional_env.unwrap_or_else(|| &temp_val_env).1
            )
        )
       ).await;
        for it in results{
            if let Err(e) = it {
                return response_error(HttpResponse::BadRequest(), e.as_str());
            }
            configurations.push(it.unwrap());
        }
    }

    let session_id  = Uuid::new_v4();

    let mut configurations = configurations;
    if let Ok(ext) = data.extensions_repo.get_all_Extension_by_extension_point(ExtensionPoint::TransformConfigurationsBeforeAttraction).await {
        configurations = ext.iter().fold(configurations.clone(), |i,r| {
            let res = r.exec(&i);
            if let Ok(z) = res {
                log::info!("Extension succeed");
                return z
            }
            if let Err(z) = res {
                log::error!("Extension error: {}", z);
            }
            i
        } );
    }

    let mut client_count: Option<i32> = None;
    if let Ok(ext) = data.extensions_repo.get_Extension_by_extension_point(ExtensionPoint::CountSessionClientsNumber).await {
        let res = ext.exec_v2(&configurations.clone());
        client_count = match res {
            Ok(x) => {
                log::info!("Extension succeed");
                Some(x)
            }
            Err(e) => {
                log::error!("Extension error: {}", e);
                None
            }
        };
    }
 
    configurations.iter().for_each(|(d,_)| log::info!("Run on: {}", d.ip));
    let configurations: Vec<_> = configurations.iter().map(|(d,c)| {
        (d,
        LaunchConfiguration{
            session_id: Some(session_id),
            need_sync: attraction_type.need_sync,
            ..c.clone()
        }
        )
    }).collect();

    let results = futures::future::join_all(
        configurations.iter()
            .map(|(d,c)| {
                run_device_configuration_inner(&data, d,c)
            })
    ).await;

    if !results.iter().all(|x| x.is_ok()) {
        for i in 0..results.len() {
            if results[i].is_err() {
                continue;
            }
            if results[i].as_ref().unwrap().0 == StatusCode::OK {
                let device_id = configurations[i].0.id;
                if let Ok((status_code, _)) = stop_device(data.clone(), session_id, device_id).await {
                    if status_code == StatusCode::OK {
                        data.device_repo.update_DeviceData_state_by_id(&device_id, DeviceState::Inactive).await;
                    }
                }
            }
        }
        return response_error(HttpResponse::InternalServerError(), "Can't run all devices")
    }

    if client_count.is_none() {
        client_count = Some(results.len() as i32);
    }

    let ids: Vec<_> =  configurations.iter().map(|(d,_)| d.id).collect();
    let _states_res = futures::future::join_all(
        ids.iter().map(|d| data.device_repo.update_DeviceData_state_by_id(&d, DeviceState::Working))
    ).await;

    let session_state = if attraction_type.need_sync == false { &SessionState::Working } else { &SessionState::Initialized };
    let session = Session::new_full(session_id,ids, &attraction_type.uid, &Utc::now(), &session_state, client_count);
    data.session_repo.add_Session(session.clone()).await;

    response_json(session)
}


#[derive(Debug, Deserialize, Serialize, Clone)]
struct SyncReq {
    callback_id: Uuid,
    session_id: Uuid
}

#[post("/runner/sync")]
async fn sync_device_configuration(req: Json<SyncReq> ,data: Data<Database>) -> HttpResponse {
    log::info!("Sync request: {:?}", req);
    let SyncReq{callback_id, session_id} = req.0;

    let SyncData{client_id,..} = data.sync_requests_repo
        .get_SyncData_by_session_and_callback_ids(&session_id, &callback_id)
        .await.unwrap();

    if let Err(_) = data.device_repo.update_DeviceData_state_by_id(&client_id, DeviceState::Prepared).await {
        return response_error(HttpResponse::InternalServerError(), "Can't update device state");
    }
    let not_synced_devices = data.sync_requests_repo.get_all_SyncData_by_session_id(&session_id);
    let count_res = data.sync_requests_repo
        .delete_SyncData_by_session_and_callback_ids(&session_id, &callback_id)
        .await;

    if let Err(_) = count_res {
        return response_error(HttpResponse::InternalServerError(), "Can't delete record from database");
    }
    let count_res = count_res.unwrap();
    if count_res == 0 {
        return response_error(HttpResponse::BadRequest(), "Invalid session_id or client_id");
    }

    let not_synced_count;
    match not_synced_devices.await {
        Err(_) => return response_error(HttpResponse::InternalServerError(), "Can't get devices list"),
        Ok(it)  => {not_synced_count = it.len();}
    }

    if not_synced_count == 0 {
        log::info!("Sync attraction");
        sync_attraction_inner(&session_id, &data).await;
        data.session_repo.update_Session_state_by_id(&session_id, SessionState::Working).await;
    }

    response_msg("Ok")
}

async fn sync_attraction_inner(session_id: &Uuid, data: &Data<Database>){
        let session = data.session_repo.get_Session_by_id(&session_id).await.unwrap();
        let devices_uids = session.devices;
        let mut devices: Vec<DeviceData> = vec![];
        for e in devices_uids.clone() {
            devices.push(data.device_repo.get_DeviceData_by_uid(&e).await.unwrap());
        }
        let attraction_type_id = session.attraction_type_id;
        let attraction_type = data.attraction_type_repo.get(&attraction_type_id).await.unwrap();
        let configurations: Vec<(Uuid, Uuid)> = attraction_type
            .post_sync_configurations
            .into_iter()
            .collect();

        for (device_type_id, configuration_id) in configurations {
            let device_type = data.device_type_repo.get(&device_type_id).await.unwrap();
            let configuration = device_type.configurations
                .iter()
                .find(|e| e.uid == configuration_id)
                .unwrap();

            let typed_devices: Vec<&DeviceData> = devices.iter()
                .filter(|e| e.device_type.clone() == device_type.uid)
                .collect();

            let mut v = Vec::new();
            for e in typed_devices {
                v.push(run_device_configuration_inner(&data, e, &configuration));
            }

            //TODO(Check result)
            let r = join_all(v).await;

        }

        let mut state_updates = Vec::new();
        devices_uids.iter().for_each(|e| {
            state_updates.push(
                data.device_repo.update_DeviceData_state_by_id(e, DeviceState::Working)
                );
        }
            );
        let r = join_all(state_updates).await;


    //TODO(return result)
}
