use std::cmp::Ordering;
use std::fs::{File, self};
use std::io;
use std::io::{BufReader, Read, Write};
use std::path::Path;
use std::process::exit;
use std::str::FromStr;
use std::time::{Duration, SystemTime};
use actix_web::{HttpResponse, web};
use actix_web::web::{Data, Json};
use filetime::FileTime;
use reqwest::blocking::Client;
use reqwest::blocking::Response;
use ssh2::Session;
use uuid::Uuid;
use crate::auxiliary::config_reader::get_exe_path;
use crate::auxiliary::response_template::{response_error, response_json, response_msg};
use crate::database_connector::Database;
use crate::model::device::DeviceState;
use crate::model::package::GitlabPackage;
use crate::model::package_file::GitlabPackageFile;
use crate::model::device_system_info::PivotData;
use crate::model::version::Version;
use crate::{obj_to_json_resp, read_config_field};
use crate::model::package_metadata::PackageMetadata;
use md5::{Md5, Digest};
use serde::{Deserialize, Serialize};


const READ_WRITE_GITLAB_REGISTRY_TOKEN: &str = "cUQVAL4yhkQxyBdDPivH";
pub const DINO_PARK_PROJECT_ID: u32 = 5;

//TODO(Fix hardcode)
pub const MANAGEMENT_SYSTEM_SOURCE_PATH: &str = "C:/Users/Developer/IdeaProjects/management_system_client/target/debug/management_system_client.exe";

pub const MANAGEMENT_SYSTEM_REMOTE_ROOT: &str = "C:/Users/management_system/";
pub const MANAGEMENT_SYSTEM_REMOTE_ROOT_CLIENT: &str = "C:/Users/management_system/management_system_client";

pub const MANAGEMENT_SYSTEM_REMOTE_PATH: &str = "C:/Users/management_system/management_system_client/management_system_client.exe";

pub const PACKAGES_FOLDER: &str = "./src/static/packages";
pub const UPDATE_SYSTEM_EXE: &str = "src/static/runner.exe";

pub fn upload_project(sess: &Session) -> Result<(), String> {
    let exe_path = Path::new(MANAGEMENT_SYSTEM_SOURCE_PATH);
    let remote_path = Path::new(MANAGEMENT_SYSTEM_REMOTE_PATH);

    let channel = sess.sftp().unwrap();

    let r_1 = channel.mkdir(Path::new(MANAGEMENT_SYSTEM_REMOTE_ROOT), 700);
    let r_2 = channel.mkdir(Path::new(MANAGEMENT_SYSTEM_REMOTE_ROOT_CLIENT), 700);
    if r_1.is_err() {
        println!("Error: {}", r_1.unwrap_err().code())
    }
    if r_2.is_err() {

        println!("Error: {}", r_2.unwrap_err().code())
    }
    let local_file = File::open(exe_path).unwrap();
    let mut f_buffer_reader = BufReader::new(local_file);

    let mut file = channel.create(remote_path).unwrap();
    let mut buf: Vec<u8> = Vec::new();
    let _ = f_buffer_reader.read_to_end(&mut buf);

    let x = file.write(buf.as_ref());
    if x.is_err() {
        return Err("Some error".to_string())
    }

    Ok(())
}

fn gitlab_api_v4_request(client: &Client, route: String, max_response_duration: Duration, token: &str) -> Result<Response, String>{
    let response_res = client
        .get(format!("https://gitlab.dreamvr.studio/api/v4/{route}"))
        .timeout(max_response_duration)
        .header("PRIVATE-TOKEN", token)
        .send();

    return if let Ok(resp) = response_res {
        Ok(resp)
    } else {
        Err("Can't get valid response".to_string())
    }
}

fn gitlab_api_v4_text_request(client: &Client, route: String, max_response_duration: Duration, token: &str) -> Result<String, String>{
    let response_res = client
        .get(format!("https://gitlab.dreamvr.studio/api/v4/{route}"))
        .timeout(max_response_duration)
        .header("PRIVATE-TOKEN", token)
        .send();

    return if let Ok(resp) = response_res {
        if let Ok(resp_text) = resp.text() {
            Ok(resp_text)
        } else {
            Err("Response isn't text".to_string())
        }
    } else {
        Err("Can't get valid response".to_string())
    }
}

fn request_packages(client: &Client, project_id: u32, max_response_duration: Duration, token: &str) -> Result<String, String>{
    gitlab_api_v4_text_request(
        client,
        format!("projects/{project_id}/packages"),
        max_response_duration,
        token
    )
}

fn request_package_files(client: &Client, project_id: u32, package_id: u32, max_response_duration: Duration, token: &str) -> Result<String, String>{
    gitlab_api_v4_text_request(
        client,
        format!("{project_id}/packages/{package_id}/package_files"),
        max_response_duration,
        token
    )
}



/// Checks the game for updates
#[get("/inner/check-for-updates/game")]
pub async fn check_for_updates(data: Data<Database>) -> HttpResponse {
    let reqwest_client = Client::new();
    let max_response_time = read_config_field("max-response-time-ms").unwrap().as_i64().unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    //TODO(Package name)
    let latest_version_future = data.software_data_repo.get_latest_version_package("sample");

    let packages_string = request_packages(&reqwest_client, DINO_PARK_PROJECT_ID, max_response_duration, READ_WRITE_GITLAB_REGISTRY_TOKEN).unwrap();
    let mut packages: Vec<GitlabPackage> = serde_json::from_str(packages_string.as_str()).unwrap();

    //Filter with tags
    packages.sort_by(|l: &GitlabPackage, r: &GitlabPackage| -> Ordering {
        r.version.cmp(&l.version)
    });

    let package_version_string = packages[0].version.to_string();
    let package_name = packages[0].name.as_str();

    let version_res = Version::from_str(package_version_string.as_str());
    if version_res.is_err() {
        return response_error(HttpResponse::InternalServerError(), "Failed to parse version: expected breaking.feat.fix")
    }

    let latest_version_res = latest_version_future.await;
    if latest_version_res.is_err() {
        return response_error(HttpResponse::InternalServerError(), "Can't retrieve data from DB")
    }
    else if let Ok(it) = latest_version_res {
        if it.version <= version_res.unwrap() {
            return obj_to_json_resp!(
                 is_new: bool = false
            )
        }
    }

    return obj_to_json_resp!(
        is_new: bool = true,
        package_data: PackageMetadata = PackageMetadata::new(package_name, version_res.unwrap())
    );
}

/// Downloads the game if needed to central server
#[get("/inner/update-if-necessary/game")]
pub async fn update_game_if_necessary(data: Data<Database>) -> HttpResponse {
    let reqwest_client = Client::new();
    let max_response_time = read_config_field("max-response-time-ms").unwrap().as_i64().unwrap() as u64;
    let max_response_duration = Duration::from_millis(max_response_time);

    //TODO(Package name)
    let latest_version_future = data.software_data_repo.get_latest_version_package("sample");

    let packages_string = request_packages(&reqwest_client, DINO_PARK_PROJECT_ID, max_response_duration, READ_WRITE_GITLAB_REGISTRY_TOKEN).unwrap();
    let mut packages: Vec<GitlabPackage> = serde_json::from_str(packages_string.as_str()).unwrap();

    //Filter with tags
    packages.sort_by(|l: &GitlabPackage, r: &GitlabPackage| -> Ordering {
        r.version.cmp(&l.version)
    });

    let package_id = packages[0].id;
    let package_version_string = packages[0].version.to_string();
    let package_name = packages[0].name.as_str();

    let version_res = Version::from_str(package_version_string.as_str());
    if version_res.is_err() {
        return response_error(HttpResponse::InternalServerError(), "Failed to parse version: expected breaking.feat.fix")
    }

    let latest_version_res = latest_version_future.await;
    if latest_version_res.is_err() {
        return response_error(HttpResponse::InternalServerError(), "Can't retrieve data from DB")
    }
    else if let Ok(it) = latest_version_res {
        if it.version <= version_res.unwrap() {
            return obj_to_json_resp!(
                 is_new: bool = false
            )
        }
    }

    let package_files = request_package_files(
        &reqwest_client, DINO_PARK_PROJECT_ID, package_id, max_response_duration, READ_WRITE_GITLAB_REGISTRY_TOKEN
    );

    let mut package_files: Vec<GitlabPackageFile> = serde_json::from_str(package_files.unwrap().as_str()).unwrap();
    package_files.sort_by(|l: &GitlabPackageFile, r: &GitlabPackageFile| -> Ordering {
        let time_l = l.created_at;
        let time_r = r.created_at;
        time_r.cmp(&time_l)
    });
    let latest_file_name = &package_files[0].file_name;

    let resp = gitlab_api_v4_request(
        &reqwest_client,
        format!("projects/{DINO_PARK_PROJECT_ID}/packages/generic/{package_name}/{package_version_string}/{latest_file_name}"),
        max_response_duration,
        READ_WRITE_GITLAB_REGISTRY_TOKEN);

    if let Err(_) = resp {
        return response_error(HttpResponse::InternalServerError(), "Can't get file")
    }
    let mut out = File::create(format!("./external/{latest_file_name}")).expect("failed to create file");
    let mut response_data = resp.unwrap();
    io::copy(&mut response_data, &mut out).expect("failed to copy content");

    let update_data_res = data.software_data_repo.update_package_version(package_name, version_res.unwrap()).await;
    if update_data_res.is_err() {
        return response_error(HttpResponse::InternalServerError(), "Can't update data")
    }

    return obj_to_json_resp!(
        is_new: bool = true,
        package_data: PackageMetadata = update_data_res.unwrap()
    );
}


#[get("/package/{id:.*}/modified")]
async fn get_package_hash(path: web::Path<(String,)>) -> HttpResponse {
    let package_id = path.0.clone();
    let package_path = format!("{}/{}", PACKAGES_FOLDER, package_id);
    let mtime = get_last_modified_time(&package_path).await;
    if let Err(e) = mtime {
        log::error!("Can't get mtime: {}", e);
        return response_error(HttpResponse::InternalServerError(), e.to_string().as_str());
    }

    let mtime = mtime.unwrap();

    obj_to_json_resp!(
        mtime_unix_seconds: i64 = mtime 
        )
}

async fn get_last_modified_time(file_path: &String) -> Result<i64, io::Error> {
   let metadata = std::fs::metadata(file_path)?; 
   Ok(FileTime::from_last_modification_time(&metadata).unix_seconds())
}

async fn get_local_md5_hash(file_path: &String) -> Result<String, io::Error> {
    let file = File::open(file_path);
    if let Err(e) = file {
        log::error!("Can't open file({}): {}", file_path, e);
        return Err(e);
    }

    let mut file = file.unwrap();
    let mut buffer = vec![];
    let read_file_res = file.read_to_end(&mut buffer);
    if let Err(e) = read_file_res {
        log::error!("Can't read file({}): {}", file_path, e);
        return Err(e);
    }

    let mut hasher = Md5::new();
    hasher.update(buffer);
    let result_hash = format!("{:x}", hasher.finalize());
    Ok(result_hash)
}


#[get("/update_all")]
async fn update_all() -> HttpResponse {
    let mut process_builder = std::process::Command::new("cmd");
    let updates_exe = format!("{}\\{}", get_exe_path(), UPDATE_SYSTEM_EXE);
    println!("Path: {}", updates_exe);
    process_builder
        .arg("/K")
        .arg(updates_exe)
        .arg("update")
        .current_dir(".\\src\\static")
        .spawn()
        .unwrap();
    exit(0);
}


#[derive(Debug, Deserialize, Serialize)]
pub struct HasUpdateReqData {
    device_id: Uuid
}


#[post("/check_update")]
pub async fn has_update(req: Json<HasUpdateReqData>, data: Data<Database>) -> HttpResponse {
    match data.device_repo.get_DeviceData_by_uid(&req.0.device_id).await {
        Ok(device) => {
            if device.device_state == DeviceState::Unavailable {
                return response_error(HttpResponse::BadRequest(), "Device is unavailable");
            }
            response_json(
                if let Ok(pkgs) = PivotData::retreive_package_updates_data(&device.host()).await {
                    pkgs.iter().any(|(name, status)| status == "Found new")
                } else {
                    false
                },
            )
        },
        Err(e) => response_error(HttpResponse::BadRequest(), e)
    }
}