use mongodb::Client;
use mongodb::options::ClientOptions;
use crate::read_config_field;
use crate::repository::attraction_type_repo::AttractionTypeRepo;
use crate::repository::device_repo::DeviceRepo;
use crate::repository::device_type_repo::DeviceTypeRepo;
use crate::repository::extensions::ExtensionsRepo;
use crate::repository::session_repo::SessionRepo;
use crate::repository::software_data_repo::SoftwareDataRepo;
use crate::repository::sync_requests_repo::SyncRequestsRepo;
use crate::repository::wol_repo::WolRepo;
use crate::repository::controller_alias_repo::ControllerAliasRepo;

const DEVICES_COLLECTION: &str = "device";
const SOFTWARE_DATA_COLLECTION: &str = "software";
const WOL_DATA_COLLECTION: &str = "wol";
const DEVICE_TYPE_COLLECTION: &str = "device_type";
const ATTRACTION_TYPE_COLLECTION: &str = "attraction_type";
const SYNC_REQUESTS_COLLECTION: &str = "sync_requests";
const SESSION_COLLECTION: &str = "session";
const EXTENSIONS_COLLECTION: &str = "extensions";
const CONTROLLER_ALIAS_COLLECTION: &str = "controller_alias";


#[derive(Clone, Debug)]
pub struct Database{
    pub client: Client,
    pub device_repo: DeviceRepo,
    pub software_data_repo: SoftwareDataRepo,
    pub wol_repo: WolRepo,
    pub device_type_repo: DeviceTypeRepo,
    pub attraction_type_repo: AttractionTypeRepo,
    pub sync_requests_repo: SyncRequestsRepo,
    pub session_repo: SessionRepo,
    pub extensions_repo: ExtensionsRepo,
    pub controller_alias_repo: ControllerAliasRepo
}

impl Database {
    pub async fn init() -> Result<Self, String> {
        let uri_yaml = read_config_field("database-uri").unwrap();
        let uri = uri_yaml.as_str().unwrap();

        let mut client_options = ClientOptions::parse(uri)
            .await
            .unwrap();

        client_options.app_name = Some("management_system".to_string());
        let client = Client::with_options(client_options).unwrap();

        let db_name_yaml = read_config_field("db-name").unwrap();
        let db_name = db_name_yaml.as_str().unwrap();

        let device_collection = client.database(db_name).collection(DEVICES_COLLECTION);
        let device_repo = DeviceRepo::new(device_collection);

        let software_data_collection = client.database(db_name).collection(SOFTWARE_DATA_COLLECTION);
        let software_data_repo = SoftwareDataRepo::new(software_data_collection);

        let wol_data_collection = client.database(db_name).collection(WOL_DATA_COLLECTION);
        let wol_repo = WolRepo::new(wol_data_collection);

        let device_type_collection = client.database(db_name).collection(DEVICE_TYPE_COLLECTION);
        let device_type_repo = DeviceTypeRepo::new(device_type_collection);

        let attraction_type_collection = client.database(db_name).collection(ATTRACTION_TYPE_COLLECTION);
        let attraction_type_repo = AttractionTypeRepo::new(attraction_type_collection);

        let sync_requests_collection = client.database(db_name).collection(SYNC_REQUESTS_COLLECTION);
        let sync_requests_repo = SyncRequestsRepo::new(sync_requests_collection).await;

        let session_collection = client.database(db_name).collection(SESSION_COLLECTION);
        let session_repo = SessionRepo::new(session_collection);

        let extensions_collection = client.database(db_name).collection(EXTENSIONS_COLLECTION);
        let extensions_repo = ExtensionsRepo::new(extensions_collection);
        
        let controller_alias_collection = client.database(db_name).collection(CONTROLLER_ALIAS_COLLECTION);
        let controller_alias_repo = ControllerAliasRepo::new(controller_alias_collection);

        Ok(Self { 
            client,
            device_repo,
            software_data_repo,
            wol_repo,
            device_type_repo,
            attraction_type_repo,
            sync_requests_repo,
            session_repo,
            extensions_repo,
            controller_alias_repo
        })
    }
}
