//#![feature(trait_alias)]
#[macro_use]
extern crate actix_web;

#[cfg(windows_service)]
#[macro_use]
extern crate windows_service;

extern crate dotenv;

#[macro_use]
extern crate dotenv_codegen;

#[cfg(windows_service)]
use windows_service::{service_control_handler, service_dispatcher};

mod api;
mod auxiliary;
mod database_connector;
mod func_tools;
mod model;
mod repository;
mod scalability_parser;

use crate::api::{admin, ads, pages, runner, updates, system};
use actix_files as fs;
use actix_web::middleware::Logger;
use actix_web::web::Data;
use actix_web::{App, HttpServer};
use api::connections;
use auxiliary::config_reader::read_config_field;
use std::io;
use std::time::Duration;
use log4rs::append::file::FileAppender;
use log4rs::Config;
use log4rs::config::{Appender, Root};
use log4rs::encode::pattern::PatternEncoder;
use warp::service;
use mac_address;
use serde::{Serialize, Deserialize};

#[cfg(windows_service)]
use windows_service::service::{
    ServiceControl, ServiceControlAccept, ServiceExitCode, ServiceState, ServiceStatus, ServiceType,
};

#[cfg(windows_service)]
use windows_service::service_control_handler::ServiceControlHandlerResult;
use crate::auxiliary::config_reader::read_ip_port;

extern crate yaml_rust;

#[cfg(windows_service)]
define_windows_service!(ffi_service_main, actix_main);

const SERVICE_NAME: &str = "mg_system";

const LICENSE_SERVER_DOMAIN: &str = "dinoserver.dreamvr.studio";

#[cfg(windows_service)]
fn run_service(arguments: Vec<OsString>) -> windows_service::Result<()> {
    let event_handler = move |control_event| -> ServiceControlHandlerResult {
        match control_event {
            ServiceControl::Stop | ServiceControl::Interrogate => {
                ServiceControlHandlerResult::NoError
            }
            _ => ServiceControlHandlerResult::NotImplemented,
        }
    };

    let service_name = SERVICE_NAME;

    // Register system service event handler
    let status_handle = service_control_handler::register(service_name, event_handler)?;

    let next_status = ServiceStatus {
        // Should match the one from system service registry
        service_type: ServiceType::OWN_PROCESS,
        // The new state
        current_state: ServiceState::Running,
        // Accept stop events when running
        controls_accepted: ServiceControlAccept::STOP,
        // Used to report an error when starting or stopping only, otherwise must be zero
        exit_code: ServiceExitCode::Win32(0),
        // Only used for pending states, otherwise must be zero
        checkpoint: 0,
        // Only used for pending states, otherwise must be zero
        wait_hint: Duration::default(),

        process_id: None,
    };

    // Tell the system that the service is running now
    status_handle.set_service_status(next_status)?;

    Ok(())
}

#[cfg(windows_service)]
fn main() -> Result<(), windows_service::Error> {
    let service_name = SERVICE_NAME;
    service_dispatcher::start(service_name, ffi_service_main)?;
    Ok(())
}

#[cfg(not(windows_service))]
#[actix_web::main]
async fn main() -> io::Result<()> {
    actix_main_inner().await
}


async fn check_license() -> Option<bool> {
    #[derive(Debug, Deserialize, Serialize)]
    struct MacAddressResponse {
        found: bool,
    }

    let mac = match mac_address::get_mac_address() {
        Ok(mac) => mac.unwrap_or_default().to_string(),
        Err(_) => String::default()
    };

    let url = format!("https://{}/check_mac_address", LICENSE_SERVER_DOMAIN);
    let payload = format!("{{\"mac\":\"{}\"}}", mac);
    log::info!("mac: {}", mac);

    let res = reqwest::Client::new()
        .post(url)
        .header("Content-Type", "application/json")
        .body(payload)
        .timeout(Duration::from_secs_f32(2.0))
        .send()
        .await;
    
    if let Ok(resp) = res {
        let res: Result<MacAddressResponse, _> = resp.json().await;
        if let Ok(data) = res {

            return Some(data.found);
        }
        return Some(false);
    }

    return None;
}


async fn actix_main_inner() -> io::Result<()> {
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    match check_license().await {
        Some(found) => {
            if !found {
                log::error!("No license");
                //std::process::exit(0);
            }
        }
        None => {
                log::error!("No license");
                //std::process::exit(0);
        }
    };

    let ip_port = read_ip_port();
    log::info!("Ip-Port: {}", ip_port);
    let database = database_connector::Database::init().await.unwrap();
    let db_data = Data::new(database);

    return HttpServer::new(move || {
        App::new()
            .app_data(db_data.clone())
            .wrap(Logger::default())
            .service(fs::Files::new("/static", "./src/static").show_files_listing())

            .service(admin::add_device_type)
            .service(admin::add_attraction_type)
            .service(admin::get_device_types)
            .service(admin::get_attractions)
            .service(admin::add_configuration)

            .service(connections::list_devices)
            .service(connections::connect)
            .service(connections::connect_via_ssh)
            .service(connections::connect_without_check)
            .service(connections::get_device_data)
            .service(connections::refresh_device_connection_state)
            .service(connections::delete_device)
            .service(connections::update_state)

            .service(pages::admin)
            .service(pages::new_device)
            .service(pages::new_attraction)
            .service(pages::run_app)
            .service(pages::run_device_with_parameters)
            .service(pages::run_attraction)
            .service(pages::get_all_devices_data)
            .service(pages::get_device_admin_page)

            .service(updates::update_game_if_necessary)
            .service(updates::check_for_updates)
            .service(updates::get_package_hash)
            .service(updates::update_all)
            .service(updates::has_update)
            .service(connections::add_wol_device)
            .service(connections::wake_device_up)
            .service(connections::wake_device_up_with_ip)
            .service(admin::add_extension)
            .service(admin::vibrate_controllers)
            .service(ads::run_twin_cat_script)
            .service(admin::update_device_state)
            .service(admin::update_device)
            .service(ads::start_ads)
            .service(ads::stop_ads)
            .service(ads::start_ads_script)
            .service(ads::stop_ads_script)

            .service(runner::run_device_configuration)
            .service(runner::run_attraction_configuration)
            .service(runner::stop_session)
            .service(runner::sync_device_configuration)
            .service(system::health_check)
            
            .service(admin::update_session_state)
            .service(admin::update_controller_alias)
            .service(admin::check_session_existence)
            .service(pages::update_admin_page)

            .wrap(Logger::new("%a %r %{User-Agent}i"))
    })
    .bind("127.0.0.1:9999")?
    .run()
    .await;
}

#[cfg(windows_service)]
#[actix_web::main]
async fn actix_main(arguments: Vec<OsString>) -> io::Result<()> {
    if let Err(_) = run_service(arguments) {}

    actix_main_inner().await
}
