use crate::to_doc_gen;
use mongodb::bson::{doc, Document};
use serde::{Deserialize, Serialize};
use std::fmt;
use std::net::IpAddr;
use std::str::FromStr;
use uuid::Uuid;
use crate::auxiliary::bson_serializable::BsonSerializable;

#[derive(Copy, Clone, Debug, Deserialize, Serialize, PartialEq)]
pub enum DeviceState {
    Inactive,
    Working,
    Prepared,
    Unavailable,
}

impl FromStr for DeviceState {
    type Err = ();

    fn from_str(input: &str) -> Result<DeviceState, Self::Err> {
        match input {
            "Inactive" => Ok(DeviceState::Inactive),
            "Working" => Ok(DeviceState::Working),
            "Prepared" => Ok(DeviceState::Prepared),
            "Unavailable" => Ok(DeviceState::Unavailable),
            _ => Err(()),
        }
    }
}

impl fmt::Display for DeviceState {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

#[derive(Clone, Debug, Deserialize, Serialize)]
pub struct DeviceData {
    pub id: Uuid,
    pub device_name: String,
    pub device_type: Uuid,
    pub device_state: DeviceState,
    pub ip: IpAddr,
    pub port: u16,
    pub mac: Option<String>
}

impl DeviceData {
    pub fn new(
        device_type: Uuid,
        device_name: Option<String>,
        ip: IpAddr,
        port: u16,
        device_state: Option<DeviceState>,
    ) -> Self {
        Self {
            id: Uuid::new_v4(),
            device_state: if device_state.is_some() {
                device_state.unwrap()
            } else {
                DeviceState::Unavailable
            },
            device_name: device_name.unwrap_or("".to_string()),
            device_type,
            ip,
            port,
            mac: None
        }
    }

    pub fn new_full(
        id: Uuid,
        device_name: Option<String>,
        device_type: Uuid,
        ip: IpAddr,
        port_: u16,
        device_state: DeviceState,
    ) -> Self {
        Self {
            id,
            device_state,
            device_type,
            device_name: device_name.unwrap_or("".to_string()),
            ip,
            port: port_,
            mac: None
        }
    }

    pub fn host(&self) -> String {
        return format!("{}:{}", self.ip, self.port);
    }
}

impl BsonSerializable for DeviceData {}
