use std::collections::HashMap;
use std::io;
use std::time::{SystemTime, Duration};
use actix_web::web::block;
use serde::{Serialize, Deserialize, de::DeserializeOwned};
use uuid::Uuid;
use crate::database_connector::Database;
use crate::model::controller_alias::ControllerAlias;
use crate::model::device_info::Process;
use super::launch_configuration::LaunchConfigurationClient;


#[derive(Debug, Deserialize, Serialize)]
pub struct SystemData{
    pub mac_addr: String,
    pub os: String,
    pub device_name: String
}

#[derive(Debug, Deserialize, Serialize)]
pub struct ControllerData{
    pub device_id: u32,
    pub charge_level: f32,
    pub serial_number: String
}

pub struct ControllerDataIncludeAlias {
    pub device_id: u32,
    pub charge_level: f32,
    pub serial_number: String,
    pub alias: String,
}

#[derive(Clone,Debug, Serialize, Deserialize)]
pub struct ProcessData {
    pub launch_parameters: LaunchConfigurationClient,
    pub uuid: Uuid,
    pub created_at: SystemTime,
}

#[derive(Clone,Debug, Serialize, Deserialize)]
pub struct ChargeLevel {
    pub levels: Vec<f32>
}

pub struct PivotData {
    pub host: String,
    pub system_data: SystemData,
    pub controllers: Vec<ControllerDataIncludeAlias>,
    pub processes: Vec<Process>,
    pub charge_level: Option<ChargeLevel>,
    pub packages: Vec<(String, String)>,
}


async fn mk_req(host: String, route: &str) -> Result<(), io::Error> {
    let url_gen = |route| {format!("http://{}{}", host, route)};

    let url = url_gen(route);

    let client = reqwest::Client::builder()
        .timeout(Duration::from_millis(500))
        .build().unwrap();
    let req = client.get(url.clone()).send().await;
    if let Err(_) = req {
        log::error!("HC error: {}", url);
        return Err(io::Error::new(io::ErrorKind::Other, "Some error"));
    }
    Ok(())
}

async fn mk_req_json<T: DeserializeOwned> (host: String, route: &str) -> Result<T, io::Error> {
    let url_gen = |route| {format!("http://{}{}", host, route)};

    let req = reqwest::get(url_gen(route)).await;
    if let Err(_) = req {
        return Err(io::Error::new(io::ErrorKind::Other, "Some error"));
    }
    let req = req.unwrap().text().await;
    if let Err(_) = req {
        return Err(io::Error::new(io::ErrorKind::Other, "Some error"));
    }
    let req = req.unwrap(); 
    let json = serde_json::from_str(&req);
    if let Err(_) = json {
        return Err(io::Error::new(io::ErrorKind::Other, "Some error"));
    }
    Ok(json.unwrap())
}


impl PivotData {

    const HC_ROUTE: &str = "/health_check"; 
    const SYSTEM_DATA_ROUTE: &str = "/system_data";
    const PROCESSES_ROUTE: &str = "/processes";
    const CONTROLLERS_ROUTE: &str = "/controllers";
    const CHARGE_LEVEL_ROUTE: &str = "/charge_level";
    const PACKAGE_UPDATES_ROUTE: &str = "/check_updates";

    pub async fn retreive_package_updates_data(host: &str) -> Result<Vec<(String, String)>, io::Error> {
        let res: Result<HashMap<String, String>, io::Error>;
        res = mk_req_json(host.to_string(), PivotData::PACKAGE_UPDATES_ROUTE).await;

        match res {
            Ok(pkgs) => {
                let mut sorted_pkgs: Vec<(String, String)> = pkgs
                    .iter()
                    .map(|(name, status)| (name.clone(), status.clone()))
                    .collect();
            
                sorted_pkgs.sort_by(|a, b| a.0.cmp(&b.0));
                Ok(sorted_pkgs)
            }
            Err(e) => Err(e)
        }
    }

    pub async fn retreive_controllers_data(host: &str, data: &actix_web::web::Data<Database>) -> Result<Vec<ControllerDataIncludeAlias>, io::Error> {
        let cs: Vec<ControllerData> = mk_req_json(host.to_string(), PivotData::CONTROLLERS_ROUTE).await?;

        // add new aliases if not exists
        for controller in &cs {
            let serial_number = &controller.serial_number;
            let controller_alias = ControllerAlias::new(serial_number.clone(), serial_number.clone());
            let res = data.controller_alias_repo
                .add_ControllerAlias(controller_alias)
                .await;
        }

        let mut cs_include_a: Vec<ControllerDataIncludeAlias> = Vec::new();
        let all_ca: Vec<ControllerAlias> = data.controller_alias_repo.get_all_ControllerAlias().await.unwrap_or(vec![]);
        for controller in cs {
            let i = all_ca.iter().position(|x| x.serial_number == controller.serial_number).unwrap();
            cs_include_a.push(ControllerDataIncludeAlias{
                device_id: controller.device_id,
                charge_level: controller.charge_level,
                serial_number: controller.serial_number,
                alias: all_ca[i].alias.clone(),
            })
        }

        Ok(cs_include_a)
    }

    pub async fn retreive_all_data(host: String, data: actix_web::web::Data<Database>) -> Result<PivotData, io::Error> {
        mk_req(host.clone(), PivotData::HC_ROUTE).await?;
        
        Ok(PivotData {
            host: host.clone(),
            system_data: mk_req_json(host.clone(), PivotData::SYSTEM_DATA_ROUTE).await?,
            processes: mk_req_json(host.clone(), PivotData::PROCESSES_ROUTE).await?,
            controllers: PivotData::retreive_controllers_data(&host, &data).await?,
            charge_level: mk_req_json(host.clone(), PivotData::CHARGE_LEVEL_ROUTE).await.ok(),
            packages: PivotData::retreive_package_updates_data(&host).await?,
        })
    }
}
