use std::collections::HashMap;
use mongodb::bson::doc;
use serde::{Deserialize, Serialize};
use actix_web::web::Data;
use uuid::Uuid;
use crate::auxiliary::bson_serializable::BsonSerializable;
use crate::database_connector::Database;
use crate::model::sync_data::SyncData;
use crate::read_ip_port;

#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct LaunchConfiguration {
    #[serde(default = "Uuid::new_v4")]
    pub uid: Uuid,
    pub session_id: Option<Uuid>,
    pub name: String,
    pub program: String,
    pub args: Vec<String>,
    pub working_dir: Option<String>,
    pub environment_vars: HashMap<String,String>,
    pub need_sync: bool,
    pub after_stop_trigger: Option<Trigger>
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct LaunchConfigurationClient {
    pub server_ip_port: String,
    pub session_id: Option<Uuid>,
    pub program: String,
    pub args: Vec<String>,
    pub working_dir: Option<String>,
    pub environment_vars: HashMap<String,String>,
    pub callback_id: Option<String>,
    pub after_stop_trigger: Option<Trigger>
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Trigger {
    pub trigger_data: String,
}

impl LaunchConfigurationClient {
   pub async fn from_launch_configuration(conf: &LaunchConfiguration, need_add_to_db: Option<(&Data<Database>, Uuid)>) -> Result<Self, ()> {
       let ip_port = read_ip_port();
       let uid: Uuid;

       if let Some((db, client_id)) = need_add_to_db{
           println!("add_sync_link");
           let sync_data = SyncData::new(client_id, &conf.session_id.unwrap());
           if let Err(_) = db.sync_requests_repo.create(sync_data.clone()).await{
               return Err(())
           }
           uid = sync_data.uid;
       }
       else {
           println!("not_sync_link");
           uid = Uuid::new_v4();
       }
       let mut args = vec![];
       args.extend(conf.args.clone());
       Ok(LaunchConfigurationClient {
           session_id: conf.session_id.clone(),
           server_ip_port: ip_port.clone(),
           program: conf.program.clone(),
           args,
           working_dir: conf.working_dir.clone(),
           environment_vars: conf.environment_vars.clone(),
           callback_id: if conf.need_sync == true {Some(uid.to_string())} else {None},
           after_stop_trigger: conf.after_stop_trigger.clone()
       })
   }
    pub fn add_arguments(&self, additional_args: &Vec<String>) -> Self {
        let mut args = vec![];
        args.extend(self.args.clone());
        args.extend(additional_args.clone());

        LaunchConfigurationClient {
            args,
            ..self.clone()
        }
    }

    pub fn add_env(&self, addition_env: &HashMap<String,String>) -> Self {
        let mut environment_vars = vec![];
        environment_vars.extend(self.environment_vars.clone());
        environment_vars.extend(addition_env.clone());
        let environment_vars = environment_vars.into_iter().collect();

        LaunchConfigurationClient {
            environment_vars,
            ..self.clone()
        }
    }
}


impl LaunchConfiguration {
    pub fn new(program: String,
               name: String,
               args: Vec<String>,
               working_dir: Option<String>,
               environment_vars: HashMap<String,String>,
               need_sync: bool,
               session_id: Option<Uuid>,
               after_stop_trigger: Option<Trigger>
    ) -> Self {
        Self {
            session_id,
            uid: Uuid::new_v4(),
            name,
            program,
            args,
            working_dir,
            environment_vars,
            need_sync,
            after_stop_trigger
        }
    }
    pub fn add_arguments(&self, additional_args: &Vec<String>) -> Self {
        let mut args = vec![];
        args.extend(self.args.clone());
        args.extend(additional_args.clone());

        Self {
            args,
            ..self.clone()
        }
    }
    pub fn add_env(&self, addition_env: &HashMap<String,String>) -> Self {
        let mut environment_vars = vec![];
        environment_vars.extend(self.environment_vars.clone());
        environment_vars.extend(addition_env.clone());
        let environment_vars = environment_vars.into_iter().collect();

        Self {
            environment_vars,
            ..self.clone()
        }
    }

}

impl BsonSerializable for LaunchConfiguration {}
