use std::fmt::{Display, Formatter};
use chrono::{DateTime, Utc};
use serde::{Serialize, Deserialize};
use uuid::Uuid;
use crate::auxiliary::bson_serializable::BsonSerializable;

#[derive(Copy, Clone, Debug, Deserialize, Serialize)]
pub enum SessionState {
    Initialized,
    Working,
    Finished,
    Aborted,
    FinishedWithError
}

impl Default for SessionState {
    fn default() -> Self {
        SessionState::Initialized
    }
}

impl Display for SessionState {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}


#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct Session {
    #[serde(default = "Uuid::new_v4")]
    pub session_id: Uuid,
    pub devices: Vec<Uuid>,
    pub attraction_type_id: Uuid,
    #[serde(with = "bson::serde_helpers::chrono_datetime_as_bson_datetime", default = "Utc::now" )]
    pub created_at: DateTime<Utc>,
    pub state: SessionState,
    pub client_count: Option<i32>
}

impl Session {
    pub fn new_full(uid: Uuid, devices : Vec<Uuid>, attraction_type_id: &Uuid, created_at: &DateTime<Utc>, state: &SessionState, client_count: Option<i32>) -> Self {
        Session {
            session_id: uid,
            devices,
            attraction_type_id: attraction_type_id.clone(),
            created_at: created_at.clone(),
            state: state.clone(),
            client_count: client_count,
        }
    }

    pub fn new(devices : Vec<Uuid>, attraction_type_id: &Uuid) -> Self {
        Session {
            session_id: Uuid::new_v4(),
            devices,
            created_at: Utc::now(),
            state: SessionState::Initialized,
            attraction_type_id: attraction_type_id.clone(),
            client_count: None
        }
    }
}

impl BsonSerializable for Session {}
