use std::cmp::Ordering;
use std::fmt;
use std::fmt::{Formatter};
use std::str::FromStr;
use serde::de::{Error, Visitor};
use serde::{Deserialize, Deserializer, Serialize, Serializer};

struct VersionVisitor;

impl<'de> Visitor<'de> for VersionVisitor {
    type Value = Version;

    fn expecting(&self, formatter: &mut Formatter) -> std::fmt::Result {
        formatter.write_str("breaking.feat.fix")
    }
    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E> where E: Error {
        let version_res = Version::from_str(v);
        return match version_res
        {
            Ok(version) => Ok(version),
            Err(_) => Err(Error::custom("Can't parse version"))
        }
    }
    fn visit_string<E>(self, v: String) -> Result<Self::Value, E> where E: Error {
        Self::visit_str(self, v.as_str())
    }
}

#[derive(Debug, Clone, Copy)]
pub struct Version{
    fix: u8,
    feat: u8,
    breaking: u8
}

impl fmt::Display for Version {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        write!(f, "{}.{}.{}", self.breaking, self.feat, self.fix)
    }
}

impl Eq for Version {}

impl PartialEq<Self> for Version {
    fn eq(&self, other: &Self) -> bool {
        self.breaking == other.breaking && self.feat == other.feat && self.fix == other.fix
    }
}

impl PartialOrd<Self> for Version {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        if self.breaking > other.breaking {
            return Some(Ordering::Greater)
        }
        else if self.breaking < other.breaking {
            return Some(Ordering::Less)
        }

        if self.feat > other.feat {
            return Some(Ordering::Greater)
        }
        else if self.feat < other.feat {
            return Some(Ordering::Less)
        }

        if self.fix > other.fix {
            return Some(Ordering::Greater)
        }
        else if self.fix < other.fix {
            return Some(Ordering::Less)
        }

        return Some(Ordering::Equal)
    }
}

impl Ord for Version {
    fn cmp(&self, other: &Self) -> Ordering {
        self.partial_cmp(other).unwrap()
    }
}

impl Serialize for Version {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
        where
            S: Serializer,
    {
        serializer.serialize_str(self.to_string().as_str())
    }
}

impl FromStr for Version {

    type Err = ();

    fn from_str(input: &str) -> Result<Version, Self::Err> {
        let version_split: Vec<&str> = input.split(".").collect();
        if version_split.len() != 3 {
            return Err(())
        }

        let breaking_res =  u8::from_str(version_split[0]);
        let feat_res = u8::from_str(version_split[1]);
        let fix_res = u8::from_str(version_split[2]);

        if breaking_res.is_err() || feat_res.is_err() || fix_res.is_err(){
            return Err(())
        }

        Ok(
            Version {
                breaking: breaking_res.unwrap(),
                feat: feat_res.unwrap(),
                fix: fix_res.unwrap(),
            }
        )
    }
}


impl<'de> Deserialize<'de> for Version {
    fn deserialize<D>(deserializer: D) -> Result<Version, D::Error>
        where
            D: Deserializer<'de>,
    {
        deserializer.deserialize_string(VersionVisitor)
    }
}