use crate::auxiliary::mongo_error::Error;
use crate::model::device_type::DeviceType;
use crate::model::launch_configuration::LaunchConfiguration;
use futures::TryStreamExt;
use mongodb::bson::{doc, Document};
use mongodb::{Collection, IndexModel};
use mongodb::options::IndexOptions;
use uuid::Uuid;
use crate::generate_body;
use crate::crud;
use crate::identify;


#[derive(Clone, Debug)]
pub struct DeviceTypeRepo {
    col: Collection<Document>,
}

impl DeviceTypeRepo {
    pub fn new(col: Collection<Document>) -> Self {
        DeviceTypeRepo { col }
    }
}

impl DeviceTypeRepo {
    crud!(get_all, DeviceType, by_name, name: &str = name.to_string());
    crud!(get, DeviceType, by_uid, uid: &Uuid = uid.to_string());
    crud!(update, DeviceType, configurations_by_uid, uid: &Uuid = uid.to_string(); configurations : &Vec<LaunchConfiguration> = bson::to_bson(configurations).unwrap());

    pub async fn create(&self, obj: DeviceType) -> Result<DeviceType, Error> {
        let _ = self
            .col
            .insert_one(obj.to_document(), None)
            .await
            .ok()
            .expect("Can't create new user");
        Ok(obj)
    }

    pub async fn get_all(&self) -> Result<Vec<DeviceType>, Error> {
        let mut cursors = self
            .col
            .find(None, None)
            .await
            .ok()
            .expect("Error getting list_devices of devices");

        let mut objs: Vec<DeviceType> = Vec::new();

        while let Some(doc) = cursors
            .try_next()
            .await
            .ok()
            .expect("Error mapping through cursor")
        {
            objs.push(DeviceType::from_document(&doc))
        }
        Ok(objs)
    }

    pub async fn get(&self, id: &Uuid) -> Result<DeviceType, &str> {
        let filter = doc! {"uid": id.to_string()};
        let mut cursors = self
            .col
            .find(filter, None)
            .await
            .ok()
            .expect("Can't get group data");

        let doc_opt = cursors
            .try_next()
            .await
            .ok()
            .expect("Error mapping through cursor");
        if doc_opt.is_none() {
            return Err("Can't unwrap device data");
        }
        let device = DeviceType::from_document(&doc_opt.unwrap());
        Ok(device)
    }


    pub async fn exists(&self, uid: Uuid) -> Result<bool, &str> {
        let filter = doc! {"uid": uid.to_string()};
        let mut cursors = self
            .col
            .find(filter, None)
            .await
            .ok()
            .expect("Can't get group data");

        let doc_opt = cursors
            .try_next()
            .await
            .ok()
            .expect("Error mapping through cursor");
        if doc_opt.is_none() {
            return Ok(false);
        }
        Ok(true)
    }


    pub async fn exists_by_name(&self, name: &str) -> Result<bool, &str> {
        let filter = doc! {"name": name};
        let mut cursors = self
            .col
            .find(filter, None)
            .await
            .ok()
            .expect("Can't get group data");

        let doc_opt = cursors
            .try_next()
            .await
            .ok()
            .expect("Error mapping through cursor");
        if doc_opt.is_none() {
            return Ok(false);
        }
        Ok(true)
    }
}
