use std::time::Duration;
use futures::TryStreamExt;
use mongodb::bson::{doc, Document};
use mongodb::{Collection, IndexModel};
use mongodb::options::IndexOptions;
use uuid::Uuid;
use crate::auxiliary::bson_serializable::BsonSerializable;
use crate::auxiliary::mongo_error::Error;
use crate::model::sync_data::SyncData;
use crate::crud;


#[derive(Clone, Debug)]
pub struct SyncRequestsRepo {
    col: Collection<Document>,
}

const INDEX_NAME: &str = "session_ttl";
const INDEX_FIELD: &str = "created_at";
const TTL: Duration = Duration::from_secs(300);

impl SyncRequestsRepo {
    pub async fn new(col: Collection<Document>) -> Self {
        let repo = SyncRequestsRepo { col };
        if !repo.check_index(INDEX_NAME.to_string()).await.unwrap_or_else(|e| -> bool{log::warn!("{}", e);false}) {
            log::info!("Can't find indexes. Creating a new one");
            repo.create_ttl_index(INDEX_NAME, INDEX_FIELD,TTL).await.unwrap();
        }
        repo
    }
}

impl SyncRequestsRepo {
    pub async fn create(&self, obj: SyncData) -> Result<SyncData, Error> {
        let _ = self
            .col
            .insert_one(obj.to_document(), None)
            .await
            .ok()
            .expect("Can't create new user");
        Ok(obj)
    }

    pub async fn get(&self, id: Uuid) -> Result<SyncData, &str> {
        let filter = doc! {"uid": id.to_string()};
        let mut cursors = self
            .col
            .find(filter, None)
            .await
            .ok()
            .expect("Can't get group data");

        let doc_opt = cursors
            .try_next()
            .await
            .ok()
            .expect("Error mapping through cursor");
        if doc_opt.is_none() {
            return Err("Can't unwrap device data");
        }
        let obj = SyncData::from_document(&doc_opt.unwrap());
        Ok(obj)
    }

    crud!(get, SyncData, by_session_and_client_ids, session_id: &Uuid = session_id.to_string(), client_id: &Uuid = client_id.to_string());
    crud!(delete, SyncData, by_session_and_client_ids, session_id: &Uuid = session_id.to_string(), client_id: &Uuid = client_id.to_string());

    crud!(get, SyncData, by_session_and_callback_ids, session_id: &Uuid = session_id.to_string(), uid: &Uuid = uid.to_string());
    crud!(delete, SyncData, by_session_and_callback_ids, session_id: &Uuid = session_id.to_string(), uid: &Uuid = uid.to_string());
    crud!(get_all, SyncData, by_session_id, session_id: &Uuid = session_id.to_string());
 

    async fn check_index(&self, index_name: String) -> Result<bool, &str> {
        match self.col.list_index_names().await {
            Ok(indexes) => {
                if let Some(_) = indexes.into_iter().find(|e| *e == index_name) {
                    return Ok(true)
                }
                Ok(false)
            }
            Err(e) => {
                log::error!("{:?}", e);
                Err("Can't get indexes")
            }
        }
    }
    async fn create_ttl_index(&self, index_name: &str, index_field: &str, ttl: Duration) -> Result<(),&str> {
        let options = IndexOptions::builder().name(index_name.to_string()).expire_after(ttl).build();
        let index = IndexModel::builder().keys(doc!{index_field: 1}).options(options).build();

        let res = self.col.create_index(index, None).await;
        if let Ok(_) = res {
            return Ok(())
        }
        Err("Can't create index")
    }
}
