use std::collections::{HashMap, HashSet};
use serde::{Deserialize, Serialize};
use uuid::Uuid;
use actix_web::web::Data;

use crate::{model::{launch_configuration::LaunchConfiguration, attraction_type::{self, AttractionType}, device::DeviceData, device_type::DeviceType}, database_connector::Database};

pub type DeviceArguments =(Uuid, Vec<String>);
pub type DeviceEnvironment = (Uuid, HashMap<String,String>);

#[derive(Debug, Deserialize, Serialize)]
pub struct AttractionReqData {
    pub attraction_id: Uuid,
    pub device_ids: Vec<Uuid>,
    pub args: Vec<DeviceArguments>,
    pub environment: Vec<DeviceEnvironment>
}

pub enum PrepareToRunError {
    DuplicateArgs,
    DuplicateEnvs,
    UnknownDeviceTypeForConfiguration {uid: Uuid},
    NotEnoughDevices {uid: Uuid},
    UnknownAttractionType,
    UnknownDevice {uid: Uuid}
}

impl AttractionReqData {
    pub async fn prepare_to_run(&self, data: Data<Database>) -> Result<Vec<LaunchConfiguration>, PrepareToRunError> {
        let attraction_type = self.retreive_attraction_type(data.clone()).await?;
        let devices = self.retreive_devices(data.clone()).await?;

        self.check_arguments()?;
        self.check_envs()?;
        self.check_minimal_set_devices(devices, attraction_type)?;

        todo!("");
    }

    fn check_arguments(&self) -> Result<(), PrepareToRunError> {
        let mut args: HashSet<Uuid> = HashSet::new(); 
        for (e,_) in &self.args {
            if args.contains(e) {
                return Err(PrepareToRunError::DuplicateArgs);
            }
            args.insert(e.clone());
        }
        Ok(())
    }

    fn check_envs(&self) -> Result<(), PrepareToRunError> {
        let mut env: HashSet<Uuid> = HashSet::new(); 
        for (e,_) in &self.environment {
            if env.contains(e) {
                return Err(PrepareToRunError::DuplicateEnvs);
            }
            env.insert(e.clone());
        }
        Ok(())
    }

    async fn retreive_attraction_type(&self, data: Data<Database>) -> Result<AttractionType, PrepareToRunError> {
        data.attraction_type_repo.get(&self.attraction_id).await
            .map_err(|_| PrepareToRunError::UnknownAttractionType)
    }

    async fn retreive_devices(&self, data: Data<Database>) -> Result<Vec<DeviceData>, PrepareToRunError> {
        let mut devices = vec![];
        for e in &self.device_ids {
            let device = data.device_repo.get_DeviceData_by_uid(e).await
                .map_err(|_| PrepareToRunError::UnknownDevice { uid: e.clone() })?;
            devices.push(device);
        }
        return Ok(devices);
    }

    async fn retreive_devices_types(&self, data: Data<Database>, attraction_type: AttractionType) -> Result<HashMap<Uuid, DeviceType>, PrepareToRunError> {
        //todo
        data.device_type_repo
    }

    fn check_minimal_set_devices(&self, devices: Vec<DeviceData>, attraction_type: AttractionType) -> Result<(), PrepareToRunError> {
        let necessary = attraction_type.minimal_necessary_devices;
        let counted = self.count_devices_by_type(devices);
        for (uid,count) in &counted {
            if !necessary.contains_key(uid) {
                return Err(PrepareToRunError::UnknownDeviceTypeForConfiguration { uid: uid.clone() });
            }
            if necessary[uid] > counted[uid] {
                return Err(PrepareToRunError::NotEnoughDevices { uid: uid.clone() })
            }
        }
        return Ok(())
    }

    fn count_devices_by_type(&self, devices: Vec<DeviceData>) -> HashMap<Uuid, i32> {
        todo!("")
    }
}
